--imports
local path = require("pl.path")
local assert = require("justo.assert")
local justo = require("justo")
local suite, test, init = justo.suite, justo.test, justo.init
local Parser = require("dogma.syn.Parser")
local SentType = require("dogma.syn.SentType")
local StmtType = require("dogma.syn.StmtType")

--Suite.
return suite("dogma.syn.StmtParser", function()
  local parser, stmt

  ----------
  -- init --
  ----------
  init("*", function()
    parser = Parser.new()
  end):title("Create parser")

  ------------------
  -- skip comment --
  ------------------
  test("skip comment", function()
    parser:parse("/**comment*/fn sum(x, y) = x + y")
    assert(parser:next()):isTable():has({type=SentType.STMT, subtype=StmtType.FN, line=1, col=13})
  end)

  ---------
  -- var --
  ---------
  suite("var", function()
    test("var x: type = value", function()
      parser:parse("var x: Type = value")
      stmt = parser:next()
      assert(stmt):isTable():has({type = SentType.VAR})
      assert(stmt.decls):len(1)
      assert(stmt.decls[1]):isTable():has({names = {"x"}, type = "var", subtype = "std"})
    end)
  end):tags("var")

  -----------
  -- break --
  -----------
  suite("break", function()
    test("break", function()
      parser:parse("break")
      stmt = parser:next()
      assert(stmt):isTable():has({type=SentType.STMT, subtype=StmtType.BREAK, loop=nil})
    end)

    test("break for", function()
      parser:parse("break for")
      stmt = parser:next()
      assert(stmt):isTable():has({type=SentType.STMT, subtype=StmtType.BREAK, loop="for"})
    end)

    test("break for each", function()
      parser:parse("break for each")
      stmt = parser:next()
      assert(stmt):isTable():has({type=SentType.STMT, subtype=StmtType.BREAK, loop="for each"})
    end)

    test("break while", function()
      parser:parse("break while")
      stmt = parser:next()
      assert(stmt):isTable():has({type=SentType.STMT, subtype=StmtType.BREAK, loop="while"})
    end)

    test("break until", function()
      parser:parse("break until")
      stmt = parser:next()
      assert(stmt):isTable():has({type=SentType.STMT, subtype=StmtType.BREAK, loop="until"})
    end)

    test("break x - error", function()
      parser:parse("break x")
      assert(function() parser:next() end):raises("anonymous code (1, 7): end of line expected.")
    end)
  end)

  ----------
  -- next --
  ----------
  suite("next", function()
    test("next", function()
      parser:parse("next")
      stmt = parser:next()
      assert(stmt):isTable():has({type = SentType.STMT, subtype = StmtType.NEXT})
    end)

    test("next x - error", function()
      parser:parse("next x")
      assert(function() parser:next() end):raises("anonymous code (1, 6): end of line expected.")
    end)
  end)

  ------------
  -- return --
  ------------
  suite("return", function()
    test("return", function()
      parser:parse("return")
      stmt = parser:next()
      assert(stmt):isTable():has({type = SentType.STMT, subtype = StmtType.RETURN})
      assert(stmt):len(0)
    end)

    test("return x+y", function()
      parser:parse("return x+y")
      stmt = parser:next()
      assert(stmt):isTable():has({type = SentType.STMT, subtype = StmtType.RETURN})
      assert(stmt):len(1)
      assert(stmt.values[1]:__tostring()):eq("(+ x y)")
    end)
  end)

  -----------
  -- yield --
  -----------
  suite("yield", function()
    test("yield x+y", function()
      parser:parse("yield x+y")
      stmt = parser:next()
      assert(stmt):isTable():has({type = SentType.STMT, subtype = StmtType.YIELD})
      assert(tostring(stmt.value)):eq("(+ x y)")
    end)
  end):tags("yield")

  ---------
  -- use --
  ---------
  suite("use", function()
    suite("template", function()
      test("use Template<T>", function()
        parser:parse('use Map<T>')
        stmt = parser:next()
        assert(stmt):isTable():has({type=SentType.STMT, subtype=StmtType.USE})
        assert(stmt.modules):eq({})
      end)

      test("use Template<InvalidName>", function()
        parser:parse("use Map<G>")
        assert(function() parser:next() end):raises("anonymous code (1, 9): name expected: 'T'. Received: 'G'.")
      end)
    end)

    suite("alias://", function()
      test("alias://id - lint", function()
        parser:parse('use alias://fs')
        stmt = parser:next()
        assert(stmt):isTable():has({type=SentType.STMT, subtype=StmtType.USE})
        assert(stmt.modules):eq({
          {name="fs", path="fs"}
        })
      end)

      test("alias://id - trans", function()
        parser:parse('use alias://fs', nil, {imports={fs="@dogmalang/fs2"}})
        stmt = parser:next()
        assert(stmt):isTable():has({type=SentType.STMT, subtype=StmtType.USE})
        assert(stmt.modules):eq({
          {name="fs", path="@dogmalang/fs2"}
        })
      end)

      test("alias://id - unknown alias", function()
        parser:parse('use alias://fs2', nil, {imports={fs="@dogmalang/fs"}})
        assert(function() parser:next() end):raises("anonymous code (1, 5): unknown module alias: fs2.")
      end)
    end)

    suite("dep:", function()
      test("dep://name", function()
        parser:parse("use dep://fs")
        stmt = parser:next()
        assert(stmt):isTable():has({type = SentType.STMT, subtype = StmtType.USE})
        assert(stmt.modules):eq({
          {name = "fs", path = "fs"}
        })
      end)

      test("dep://name-name", function()
        parser:parse("use dep://fs-extra")
        stmt = parser:next()
        assert(stmt):isTable():has({type = SentType.STMT, subtype = StmtType.USE})
        assert(stmt.modules):eq({
          {name = "extra", path = "fs-extra"}
        })
      end)

      test("dep://name/name", function()
        parser:parse("use dep://mypkg/mymod")
        stmt = parser:next()
        assert(stmt):isTable():has({type = SentType.STMT, subtype = StmtType.USE})
        assert(stmt.modules):eq({
          {name = "mymod", path = "mypkg/mymod"}
        })
      end)

      test("dep://name/name-name", function()
        parser:parse("use dep://mypkg/my-mod")
        stmt = parser:next()
        assert(stmt):isTable():has({type = SentType.STMT, subtype = StmtType.USE})
        assert(stmt.modules):eq({
          {name = "mod", path = "mypkg/my-mod"}
        })
      end)

      test("dep://name-name/name-name", function()
        parser:parse("use dep://my-pkg/my-mod")
        stmt = parser:next()
        assert(stmt):isTable():has({type = SentType.STMT, subtype = StmtType.USE})
        assert(stmt.modules):eq({
          {name = "mod", path = "my-pkg/my-mod"}
        })
      end)
    end)

    test("use LiteralStr", function()
      parser:parse('use "module"')
      stmt = parser:next()
      assert(stmt):isTable():has({type = SentType.STMT, subtype = StmtType.USE})
      assert(stmt.modules):eq({
        {name = "module", path = "module"}
      })
    end)

    test("use\\n LiteralStr", function()
      parser:parse('use\n "module"')
      stmt = parser:next()
      assert(stmt):isTable():has({type = SentType.STMT, subtype = StmtType.USE})
      assert(stmt.modules):eq({
        {name = "module", path = "module"}
      })
    end)

    test("use\\nLiteralStr", function()
      parser:parse('use\n"module"')
      stmt = parser:next()
      assert(stmt):isTable():has({type = SentType.STMT, subtype = StmtType.USE})
      assert(stmt.modules):eq({})
    end)

    test('use "justo.assert"', function()
      parser:parse('use "justo.assert"')
      stmt = parser:next()
      assert(stmt):isTable():has({type = SentType.STMT, subtype = StmtType.USE})
      assert(stmt.modules):eq({
        {name = "assert", path = "justo.assert"}
      })
    end)

    test('use "redispark-connector-redis"', function()
      parser:parse('use "redispark-connector-redis"')
      stmt = parser:next()
      assert(stmt):isTable():has({type = SentType.STMT, subtype = StmtType.USE})
      assert(stmt.modules):eq({
        {name = "redis", path = "redispark-connector-redis"}
      })
    end)

    test('use "@scope/pkg"', function()
      parser:parse('use "@scope/pkg"')
      stmt = parser:next()
      assert(stmt):isTable():has({type = SentType.STMT, subtype = StmtType.USE})
      assert(stmt.modules):eq({
        {name = "pkg", path = "@scope/pkg"}
      })
    end)

    test('use "@scope/my-pkg"', function()
      parser:parse('use "@scope/my-pkg"')
      stmt = parser:next()
      assert(stmt):isTable():has({type = SentType.STMT, subtype = StmtType.USE})
      assert(stmt.modules):eq({
        {name = "pkg", path = "@scope/my-pkg"}
      })
    end)

    test("use LitealStr, LiteralStr", function()
      parser:parse('use "module1", "module2"')
      stmt = parser:next()
      assert(stmt):isTable():has({type = SentType.STMT, subtype = StmtType.USE})
      assert(stmt.modules):eq({
        {name = "module1", path = "module1"},
        {name = "module2", path = "module2"}
      })
    end)

    test("use\\n  LiteralStr\\n  LiteralStr", function()
      parser:parse('use\n "module1"\n "module2"')
      stmt = parser:next()
      assert(stmt):isTable():has({type = SentType.STMT, subtype = StmtType.USE})
      assert(stmt.modules):eq({
        {name = "module1", path = "module1"},
        {name = "module2", path = "module2"}
      })
    end)

    test("use LiteralStr as Name", function()
      parser:parse('use "module" as mod')
      stmt = parser:next()
      assert(stmt):isTable():has({type = SentType.STMT, subtype = StmtType.USE})
      assert(stmt.modules):eq({
        {name = "mod", path = "module"}
      })
    end)

    test("use\\n LiteralStr as Name", function()
      parser:parse('use\n "module" as mod')
      stmt = parser:next()
      assert(stmt):isTable():has({type = SentType.STMT, subtype = StmtType.USE})
      assert(stmt.modules):eq({
        {name = "mod", path = "module"}
      })
    end)

    test("use LiteralStr as Name, LiteralStr as Name", function()
      parser:parse('use "module1" as mod1, "module2" as mod2')
      stmt = parser:next()
      assert(stmt):isTable():has({type = SentType.STMT, subtype = StmtType.USE})
      assert(stmt.modules):eq({
        {name = "mod1", path = "module1"},
        {name = "mod2", path = "module2"}
      })
    end)

    test("use\\n LiteralStr as Name\\n LiteralStr as Name", function()
      parser:parse('use\n "module1" as mod1\n "module2" as mod2')
      stmt = parser:next()
      assert(stmt):isTable():has({type = SentType.STMT, subtype = StmtType.USE})
      assert(stmt.modules):eq({
        {name = "mod1", path = "module1"},
        {name = "mod2", path = "module2"}
      })
    end)

    test([[use "module1"! "module2" - error - comma expected]], function()
      parser:parse([[use "module1"! "module2"]])
      assert(function() parser:next() end):raises("anonymous code (1, 14): ',' or end of line expected.")
    end)

    test([[use "~module" - error - invalid module path format]], function()
      parser:parse([[use "~module"]])
      assert(function() parser:next() end):raises("invalid module path format: '~module'.")
    end)

    test([[use "one/two"]], function()
      parser:parse([[use "one/two"]])
      stmt = parser:next()
      assert(stmt):isTable():has({type = SentType.STMT, subtype = StmtType.USE})
      assert(stmt.modules):eq({
        {name = "two", path = "one/two"}
      })
    end)

    test("use Literal, Name", function()
      parser:parse('use "one", two')
      stmt = parser:next()
      assert(stmt):isTable():has({type = SentType.STMT, subtype = StmtType.USE})
      assert(stmt.modules):eq({
        {name = "one", path = "one"},
        {name = "two", path = "./two"}
      })
    end)

    suite("internal", function()
      test("use name", function()
        parser:parse("use file")
        stmt = parser:next()
        assert(stmt):isTable():has({type=SentType.STMT, subtype=StmtType.USE})
        assert(stmt.modules):eq({
          {name="file", path="./file"}
        })
      end)

      test("use name/", function()
        parser:parse("use my/dir/")
        stmt = parser:next()
        assert(stmt):isTable():has({type=SentType.STMT, subtype=StmtType.USE})
        assert(stmt.modules):eq({
          {path="./my/dir/"}
        })
      end)

      test("use name/{name, name, name}", function()
        parser:parse("use dir/{One, Two, Three}")
        stmt = parser:next()
        assert(stmt):isTable():has({type=SentType.STMT, subtype=StmtType.USE})
        assert(stmt.modules):eq({
          {name = "One", path = "./dir/One"},
          {name = "Two", path = "./dir/Two"},
          {name = "Three", path = "./dir/Three"}
        })
      end)

      test("use {uVars} = name", function()
        parser:parse("use {var1, var2 as VAR2, var3} = module")
        stmt = parser:next()
        assert(stmt):isTable():has({type=SentType.STMT, subtype=StmtType.USE})
        assert(stmt.modules):eq({
          {name=nil, path="./module", uVars={"var1", {name = "var2", as = "VAR2"}, "var3"}}
        })
      end)

      test("use Alias = Name", function()
        parser:parse('use mod1 = module1')
        stmt = parser:next()
        assert(stmt):isTable():has({type = SentType.STMT, subtype = StmtType.USE})
        assert(stmt.modules):eq({
          {name = "mod1", path = "./module1"},
        })
      end)

      test("use name/name", function()
        parser:parse("use dir/file")
        stmt = parser:next()
        assert(stmt):isTable():has({type=SentType.STMT, subtype=StmtType.USE})
        assert(stmt.modules):eq({
          {name="file", path="./dir/file"}
        })
      end)

      test("use name/name/name", function()
        parser:parse("use dir/subdir/file")
        stmt = parser:next()
        assert(stmt):isTable():has({type=SentType.STMT, subtype=StmtType.USE})
        assert(stmt.modules):eq({
          {name="file", path="./dir/subdir/file"}
        })
      end)

      test("use ../name", function()
        parser:parse("use ../file")
        stmt = parser:next()
        assert(stmt):isTable():has({type=SentType.STMT, subtype=StmtType.USE})
        assert(stmt.modules):eq({
          {name="file", path="../file"}
        })
      end)

      test("use ../name/{name, name, name}", function()
        parser:parse("use ../dir/{One, Two, Three}")
        stmt = parser:next()
        assert(stmt):isTable():has({type=SentType.STMT, subtype=StmtType.USE})
        assert(stmt.modules):eq({
          {name = "One", path = "../dir/One"},
          {name = "Two", path = "../dir/Two"},
          {name = "Three", path = "../dir/Three"}
        })
      end)

      test("use ../name/name", function()
        parser:parse("use ../dir/file")
        stmt = parser:next()
        assert(stmt):isTable():has({type=SentType.STMT, subtype=StmtType.USE})
        assert(stmt.modules):eq({
          {name="file", path="../dir/file"}
        })
      end)

      test("use ../name/name/", function()
        parser:parse("use ../my/dir/")
        stmt = parser:next()
        assert(stmt):isTable():has({type=SentType.STMT, subtype=StmtType.USE})
        assert(stmt.modules):eq({
          {path="../my/dir/"}
        })
      end)

      test("use ../../name", function()
        parser:parse("use ../../file")
        stmt = parser:next()
        assert(stmt):isTable():has({type=SentType.STMT, subtype=StmtType.USE})
        assert(stmt.modules):eq({
          {name="file", path="../../file"}
        })
      end)

      test("use ../../name/name", function()
        parser:parse("use ../../dir/file")
        stmt = parser:next()
        assert(stmt):isTable():has({type=SentType.STMT, subtype=StmtType.USE})
        assert(stmt.modules):eq({
          {name="file", path="../../dir/file"}
        })
      end)

      test("use ~/file", function()
        parser:parse("use ~/file", path.join(path.currentdir(), "one/two/three/script.dog"))
        stmt = parser:next()
        assert(stmt):isTable():has({type=SentType.STMT, subtype=StmtType.USE})
        assert(stmt.modules):eq({
          {name="file", path="../../../file"}
        })
      end)

      test("use ~/dir/file", function()
        parser:parse("use ~/my/file", path.join(path.currentdir(), "one/two/three/file.dog"))
        stmt = parser:next()
        assert(stmt):isTable():has({type=SentType.STMT, subtype=StmtType.USE})
        assert(stmt.modules):eq({
          {name="file", path="../../../my/file"}
        })
      end)

      test("use (\\nname1, name2\\nname3)", function()
        parser:parse("use (\nname1, name2\nname3\n)")
        stmt = parser:next()
        assert(stmt):isTable():has({type=SentType.STMT, subtype=StmtType.USE})
        assert(stmt.modules):eq({
          {name="name1", path="./name1"},
          {name="name2", path="./name2"},
          {name="name3", path="./name3"}
        })
      end)
    end)

    test("use Number - error", function()
      parser:parse("use 123")
      assert(function() parser:next() end):raises("anonymous code (1, 5): 'alias://', 'dep://', name, literal text, '~' or '..' expected.")
    end)
  end):tags("use")

  ----------
  -- from --
  ----------
  suite("from", function()
    test("from ~/my/file", function()
      parser:parse([[from ~/my/file use FSWatcher]], path.join(path.currentdir(), "one/two/script.dog"))
      stmt = parser:next()
      assert(stmt):isTable():has({
        type = SentType.STMT,
        subtype = StmtType.FROM,
        line = 1,
        col = 1,
        module = "../../my/file",
        members = {
          {name = "FSWatcher", as = "FSWatcher"}
        }
      })
    end)

    test("from LiteralStr use Name", function()
      parser:parse([[from "fs" use FSWatcher]])
      stmt = parser:next()
      assert(stmt):isTable():has({
        type = SentType.STMT,
        subtype = StmtType.FROM,
        line = 1,
        col = 1,
        module = "fs",
        members = {
          {name = "FSWatcher", as = "FSWatcher"}
        }
      })
    end)

    test("from LiteralStr use Name as Name", function()
      parser:parse([[from "fs" use FSWatcher as Watcher]])
      stmt = parser:next()
      assert(stmt):isTable():has({
        type = SentType.STMT,
        subtype = StmtType.FROM,
        line = 1,
        col = 1,
        module = "fs",
        members = {
          {name = "FSWatcher", as = "Watcher"}
        }
      })
    end)

    test("from LiteralStr use Name, Name", function()
      parser:parse([[from "fs" use FSWatcher, watch]])
      stmt = parser:next()
      assert(stmt):isTable():has({
        type = SentType.STMT,
        subtype = StmtType.FROM,
        line = 1,
        col = 1,
        module = "fs",
        members = {
          {name = "FSWatcher", as = "FSWatcher"},
          {name = "watch", as = "watch"}
        }
      })
    end)

    test("from LiteralStr use Name as Name, Name as Name", function()
      parser:parse([[from "fs" use FSWatcher as Watcher, watch as wtch]])
      stmt = parser:next()
      assert(stmt):isTable():has({
        type = SentType.STMT,
        subtype = StmtType.FROM,
        line = 1,
        col = 1,
        module = "fs",
        members = {
          {name = "FSWatcher", as = "Watcher"},
          {name = "watch", as = "wtch"}
        }
      })
    end)

    test("from LiteralStr use (\\n  Name as Name\\n  Name as Name\n)", function(params)
      parser:parse(params[1])
      assert(parser:next()):isTable():has({
        type = SentType.STMT,
        subtype = StmtType.FROM,
        line = 1,
        col = 1,
        module = "fs",
        members = {
          {name = "FSWatcher", as = "Watcher"},
          {name = "watch", as = "wtch"}
        }
      })
      assert(parser:next()):isNil()
    end):iter(
      {subtitle=" # newline as separator", params='from "fs" use (\n  FSWatcher as Watcher\n  watch as wtch\n)'},
      {subtitle=" # comma as separator", params='from "fs" use (\n  FSWatcher as Watcher,\n  watch as wtch\n)'}
    )

    test("from LiteralStr use (\\n  Name as Name\\n  Name as Name)", function()
      parser:parse('from "fs" use (\n  FSWatcher as Watcher\n  watch as wtch)')
      assert(parser:next()):isTable():has({
        type = SentType.STMT,
        subtype = StmtType.FROM,
        line = 1,
        col = 1,
        module = "fs",
        members = {
          {name = "FSWatcher", as = "Watcher"},
          {name = "watch", as = "wtch"}
        }
      })
      assert(parser:next()):isNil()
    end)

    test("from LiteralStr use Name => {variables}", function()
      parser:parse('from "module" use mem1 as MEM1 => {var1,var2}, mem2 => {var3,var4}\nabc')
      assert(parser:next()):isTable():has({
        type = SentType.STMT,
        subtype = StmtType.FROM,
        line = 1,
        col = 1,
        module = "module",
        members = {
          {name="mem1", as="MEM1", uVars={"var1","var2"}},
          {name="mem2", as="mem2", uVars={"var3","var4"}}
        }
      })
      assert(parser:next()):isTable():has({
        type = SentType.EXP,
        line = 2,
        col = 1
      })
    end)

    suite("errors", function()
      test("name expected after comma", function()
        parser:parse('from "fs" use\n  FSWatcher as Watcher,\n  watch as wtch')
        assert(function() parser:next() end):raises("anonymous code (1, 14): name expected.")
      end)
    end)
  end):tags("from")

  ----------
  -- enum --
  ----------
  suite("enum", function()
    test("enum - error", function()
      parser:parse("enum")
      assert(function() parser:next() end):raises("anonymous code (1, 5): name expected.")
    end)

    test("visibility enum Color", function(params)
      parser:parse(params[1] .. " enum Color")
      stmt = parser:next()
      assert(stmt):isTable():has({
        type = SentType.STMT,
        subtype = StmtType.ENUM,
        visib = params[1],
        name = "Color",
        items = {}
      })
    end):iter(
      {subtitle = " # export", params = "export"},
      {subtitle = " # pub", params = "pub"}
    )

    test("enum Color", function()
      parser:parse("enum Color")
      stmt = parser:next()
      assert(stmt):isTable():has({
        type = SentType.STMT,
        subtype = StmtType.ENUM,
        visib = nil,
        name = "Color",
        items = {}
      })
    end)

    test("enum Color RED - error", function()
      parser:parse("enum Color RED")
      assert(function() parser:next() end):raises("anonymous code (1, 12): invalid token.")
    end)

    test("visibility enum Color", function(params)
      parser:parse(params[1] .. " enum Color")
      stmt = parser:next()
      assert(stmt):isTable():has({
        type = SentType.STMT,
        subtype = StmtType.ENUM,
        visib = params[1],
        name = "Color",
        items = {}
      })
    end):iter(
      {subtitle = "# export", params = "export"},
      {subtitle = "# pub", params = "pub"}
    )

    test("pub enum Color", function()
      parser:parse("pub enum Color")
      stmt = parser:next()
      assert(stmt):isTable():has({
        type = SentType.STMT,
        subtype = StmtType.ENUM,
        visib = "pub",
        name = "Color",
        items = {}
      })
    end)

    test("enum Color {RED}", function()
      parser:parse("enum Color {RED}")
      stmt = parser:next()
      assert(stmt):isTable():has({
        type = SentType.STMT,
        subtype = StmtType.ENUM,
        name = "Color",
        items = {{name = "RED", value = 1}}
      })
    end)

    test("enum Color {RED, GREEN}", function()
      parser:parse("enum Color {RED, GREEN}")
      stmt = parser:next()
      assert(stmt):isTable():has({
        type = SentType.STMT,
        subtype = StmtType.ENUM,
        name = "Color",
        items = {{name = "RED", value = 1}, {name = "GREEN", value = 2}}
      })
    end)

    test("enum Color {RED, GREEN, BLUE}", function()
      parser:parse("enum Color {RED, GREEN, BLUE}")
      stmt = parser:next()
      assert(stmt):isTable():has({
        type = SentType.STMT,
        subtype = StmtType.ENUM,
        name = "Color",
        items = {
          {name = "RED", value = 1},
          {name = "GREEN", value = 2},
          {name = "BLUE", value = 3}
        }
      })
    end)

    test("enum Color {RED = 12}", function()
      parser:parse("enum Color {RED = 12}")
      stmt = parser:next()
      assert(stmt):isTable():has({
        type = SentType.STMT,
        subtype = StmtType.ENUM,
        name = "Color",
        items = {{name = "RED", value = 12}}
      })
    end)

    test("enum Color {RED = 12, GREEN = 34}", function()
      parser:parse("enum Color {RED = 12, GREEN = 34}")
      stmt = parser:next()
      assert(stmt):isTable():has({
        type = SentType.STMT,
        subtype = StmtType.ENUM,
        name = "Color",
        items = {{name = "RED", value = 12}, {name = "GREEN", value = 34}}
      })
    end)

    test("enum Color {RED = 12, GREEN = 34, BLUE = 56}", function()
      parser:parse("enum Color {RED = 12, GREEN = 34, BLUE = 56}")
      stmt = parser:next()
      assert(stmt):isTable():has({
        type = SentType.STMT,
        subtype = StmtType.ENUM,
        name = "Color",
        items = {
          {name = "RED", value = 12},
          {name = "GREEN", value = 34},
          {name = "BLUE", value = 56}
        }
      })
    end)

    test("enum Color\\n  RED", function()
      parser:parse("enum Color\n  RED")
      stmt = parser:next()
      assert(stmt):isTable():has({
        type = SentType.STMT,
        subtype = StmtType.ENUM,
        name = "Color",
        items = {{name = "RED", value = 1}}
      })
    end)

    test("enum Color\\nRED", function()
      parser:parse("enum Color\nRED")
      stmt = parser:next()
      assert(stmt):isTable():has({
        type = SentType.STMT,
        subtype = StmtType.ENUM,
        name = "Color",
        items = {}
      })
    end)

    test("enum Color\\n  RED\\n  GREEN", function()
      parser:parse("enum Color\n  RED\n  GREEN")
      stmt = parser:next()
      assert(stmt):isTable():has({
        type = SentType.STMT,
        subtype = StmtType.ENUM,
        name = "Color",
        items = {{name = "RED", value = 1}, {name = "GREEN", value = 2}}
      })
    end)

    test("enum Color\\n  RED\\nGREEN", function()
      parser:parse("enum Color\n  RED\nGREEN")
      stmt = parser:next()
      assert(stmt):isTable():has({
        type = SentType.STMT,
        subtype = StmtType.ENUM,
        name = "Color",
        items = {{name = "RED", value = 1}}
      })
    end)

    test("enum Color\\n  RED\\n  GREEN\\n  BLUE", function()
      parser:parse("enum Color\n  RED\n  GREEN\n  BLUE")
      stmt = parser:next()
      assert(stmt):isTable():has({
        type = SentType.STMT,
        subtype = StmtType.ENUM,
        name = "Color",
        items = {
          {name = "RED", value = 1},
          {name = "GREEN", value = 2},
          {name = "BLUE", value = 3}
        }
      })
    end)

    test("enum Color\\n  RED = 12", function()
      parser:parse("enum Color\n  RED = 12")
      stmt = parser:next()
      assert(stmt):isTable():has({
        type = SentType.STMT,
        subtype = StmtType.ENUM,
        name = "Color",
        items = {{name = "RED", value = 12}}
      })
    end)

    test("enum Color\\n  RED = 12\\n  GREEN = 34", function()
      parser:parse("enum Color\n  RED = 12\n  GREEN = 34")
      stmt = parser:next()
      assert(stmt):isTable():has({
        type = SentType.STMT,
        subtype = StmtType.ENUM,
        name = "Color",
        items = {{name = "RED", value = 12}, {name = "GREEN", value = 34}}
      })
    end)

    test("enum Color\\n  RED = 12\\n  GREEN = 34\\n  BLUE = 56", function()
      parser:parse("enum Color\n  RED = 12\n  GREEN = 34\n  BLUE = 56")
      stmt = parser:next()
      assert(stmt):isTable():has({
        type = SentType.STMT,
        subtype = StmtType.ENUM,
        name = "Color",
        items = {
          {name = "RED", value = 12},
          {name = "GREEN", value = 34},
          {name = "BLUE", value = 56}
        }
      })
    end)

    test("@annot1 @annot2\\nenum Color\\n  RED = 12\\n  GREEN = 34\\n  BLUE = 56", function()
      parser:parse("@annot1 @annot2\nenum Color\n  RED = 12\n  GREEN = 34\n  BLUE = 56")
      stmt = parser:next()
      assert(stmt):isTable():has({
        type = SentType.STMT,
        subtype = StmtType.ENUM,
        name = "Color",
        items = {
          {name = "RED", value = 12},
          {name = "GREEN", value = 34},
          {name = "BLUE", value = 56}
        }
      })
      assert(stmt.annots:has("annot1")):eq(true)
      assert(stmt.annots:has("annot2")):eq(true)
    end)

    test("@annot1\\n@annot2\\nenum Color\\n  RED = 12\\n  GREEN = 34\\n  BLUE = 56", function()
      parser:parse("@annot1\n@annot2\nenum Color\n  RED = 12\n  GREEN = 34\n  BLUE = 56")
      stmt = parser:next()
      assert(stmt):isTable():has({
        type = SentType.STMT,
        subtype = StmtType.ENUM,
        name = "Color",
        items = {
          {name = "RED", value = 12},
          {name = "GREEN", value = 34},
          {name = "BLUE", value = 56}
        }
      })
      assert(stmt.annots:has("annot1")):eq(true)
      assert(stmt.annots:has("annot2")):eq(true)
    end)

    test("enum with comments whitin elements", function()
      parser:parse("enum Color\n  #comment\nRED = 12\n\n#comment\n  GREEN = 34\n  BLUE = 56")
      stmt = parser:next()
      assert(stmt):isTable():has({
        type = SentType.STMT,
        subtype = StmtType.ENUM,
        name = "Color",
        items = {
          {name = "RED", value = 12},
          {name = "GREEN", value = 34},
          {name = "BLUE", value = 56}
        }
      })
    end)
  end):tags("enum")

  -----------
  -- while --
  -----------
  suite("while", function()
    test("while Exp do Exp", function()
      parser:parse("while true do print(x)")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        type = SentType.STMT,
        subtype = StmtType.WHILE,
        decl = nil,
        value = nil
      })
      assert(stmt.cond:__tostring()):eq("true")
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(call print x)")
      assert(stmt.catch):isNil()
      assert(stmt.finally):isNil()

      assert(parser:next()):isNil()
    end)

    test("while decl do Exp", function()
      parser:parse("while x := 1 do print(x)")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        type = SentType.STMT,
        subtype = StmtType.WHILE,
        decl = "var"
      })
      assert(stmt.value:__tostring()):eq("(= x 1)")
      assert(stmt.cond):isNil()
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(call print x)")
      assert(stmt.catch):isNil()
      assert(stmt.finally):isNil()

      assert(parser:next()):isNil()
    end)

    test("while decl; Exp do Exp", function()
      parser:parse("while x := 1; x > 10 do print(x)")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        type = SentType.STMT,
        subtype = StmtType.WHILE,
        decl = "var"
      })
      assert(stmt.value:__tostring()):eq("(= x 1)")
      assert(stmt.cond:__tostring()):eq("(> x 10)")
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(call print x)")
      assert(stmt.catch):isNil()
      assert(stmt.finally):isNil()

      assert(parser:next()):isNil()
    end)

    test("while Exp do", function()
      parser:parse("while true do")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        type = SentType.STMT,
        subtype = StmtType.WHILE,
        decl = nil,
        value = nil
      })
      assert(stmt.cond:__tostring()):eq("true")
      assert(stmt.body):isEmpty()
      assert(stmt.catch):isNil()
      assert(stmt.finally):isNil()

      assert(parser:next()):isNil()
    end)

    test("while Exp do\\n Sentence", function()
      parser:parse("while true do\n print(x)")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        type = SentType.STMT,
        subtype = StmtType.WHILE,
        decl = nil,
        value = nil
      })
      assert(stmt.cond:__tostring()):eq("true")
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(call print x)")
      assert(stmt.catch):isNil()
      assert(stmt.finally):isNil()

      assert(parser:next()):isNil()
    end)

    test("while decl; Exp do\\n Sentence", function()
      parser:parse("while x ::= 1; x < 100 do\n print(x)")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        type = SentType.STMT,
        subtype = StmtType.WHILE,
        decl = "const"
      })
      assert(stmt.value:__tostring()):eq("(= x 1)")
      assert(stmt.cond:__tostring()):eq("(< x 100)")
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(call print x)")
      assert(stmt.catch):isNil()
      assert(stmt.finally):isNil()

      assert(parser:next()):isNil()
    end)

    test("while Exp do\\n Sentence\\n Sentence", function()
      parser:parse("while true do\n print(x)\n print(y)")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        type = SentType.STMT,
        subtype = StmtType.WHILE,
        decl = nil,
        value = nil
      })
      assert(stmt.cond:__tostring()):eq("true")
      assert(stmt.body):len(2)
      assert(stmt.body[1]:__tostring()):eq("(call print x)")
      assert(stmt.body[2]:__tostring()):eq("(call print y)")
      assert(stmt.catch):isNil()
      assert(stmt.finally):isNil()

      assert(parser:next()):isNil()
    end)

    test("while Exp do\\n Sentence\\nSentence", function()
      parser:parse("while true do\n print(x)\nprint(y)")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        type = SentType.STMT,
        subtype = StmtType.WHILE,
        decl = nil,
        value = nil
      })
      assert(stmt.cond:__tostring()):eq("true")
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(call print x)")
      assert(stmt.catch):isNil()
      assert(stmt.finally):isNil()

      assert(parser:next():__tostring()):eq("(call print y)")
      assert(parser:next()):isNil()
    end)

    test("while Exp do Block else Block", function()
      parser:parse("while true do\n print(x)\nelse\n print(y)")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        type = SentType.STMT,
        subtype = StmtType.WHILE,
        decl = nil,
        value = nil
      })
      assert(stmt.cond:__tostring()):eq("true")
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(call print x)")
      assert(stmt.el):len(1)
      assert(stmt.el[1]:__tostring()):eq("(call print y)")
      assert(stmt.catch):isNil()
      assert(stmt.finally):isNil()

      assert(parser:next()):isNil()
    end)

    test("while Exp do\\n Sentence\\ncatch\\n  Sentence", function()
      parser:parse("while true do\n print(x)\ncatch\n  print(e)")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        type = SentType.STMT,
        subtype = StmtType.WHILE,
        decl = nil,
        value = nil
      })
      assert(stmt.cond:__tostring()):eq("true")
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(call print x)")
      assert(stmt.catch):isTable()
      assert(stmt.catch.var):isNil()
      assert(stmt.catch.body):len(1)
      assert(stmt.catch.body[1]:__tostring()):eq("(call print e)")
      assert(stmt.finally):isNil()

      assert(parser:next()):isNil()
    end)

    test("while Exp do\\n print(x)\\ncatch\\nprint(e)", function()
      parser:parse("while true do\n print(x)\ncatch\nprint(e)")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        type = SentType.STMT,
        subtype = StmtType.WHILE,
        decl = nil,
        value = nil
      })
      assert(stmt.cond:__tostring()):eq("true")
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(call print x)")
      assert(stmt.catch):isTable()
      assert(stmt.catch.var):isNil()
      assert(stmt.catch.body):len(0)
      assert(stmt.finally):isNil()

      assert(parser:next():__tostring()):eq("(call print e)")
      assert(parser:next()):isNil()
    end)

    test("while Exp do\\n print(x)\\ncatch e\\n  print(e)", function()
      parser:parse("while true do\n  print(x)\ncatch e\n  print(e)")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        type = SentType.STMT,
        subtype = StmtType.WHILE,
        decl = nil,
        value = nil
      })
      assert(stmt.cond:__tostring()):eq("true")
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(call print x)")
      assert(stmt.catch):isTable()
      assert(stmt.catch.var):eq("e")
      assert(stmt.catch.body):len(1)
      assert(stmt.catch.body[1]:__tostring()):eq("(call print e)")
      assert(stmt.finally):isNil()

      assert(parser:next()):isNil()
    end)

    test("while Exp do\\n print(x)\\nfinally\\n  print(y)", function()
      parser:parse("while true do\n print(x)\nfinally\n print(y)")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        type = SentType.STMT,
        subtype = StmtType.WHILE,
        decl = nil,
        value = nil
      })
      assert(stmt.cond:__tostring()):eq("true")
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(call print x)")
      assert(stmt.catch):isNil()
      assert(stmt.finally):isTable()
      assert(stmt.finally.body):len(1)
      assert(stmt.finally.body[1]:__tostring()):eq("(call print y)")

      assert(parser:next()):isNil()
    end)

    test("while Exp do\\n print(x)\\ncatch e\\n print(y)\\nfinally\\n print(z)", function()
      parser:parse("while true do\n print(x)\ncatch e\n print(y)\nfinally\n print(z)")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        type = SentType.STMT,
        subtype = StmtType.WHILE,
        decl = nil,
        value = nil
      })
      assert(stmt.cond:__tostring()):eq("true")
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(call print x)")
      assert(stmt.catch):isTable()
      assert(stmt.catch.var):eq("e")
      assert(stmt.catch.body):len(1)
      assert(stmt.catch.body[1]:__tostring()):eq("(call print y)")
      assert(stmt.finally):isTable()
      assert(stmt.finally.body):len(1)
      assert(stmt.finally.body[1]:__tostring()):eq("(call print z)")

      assert(parser:next()):isNil()
    end)
  end):tags("while")

  -----------
  -- until --
  -----------
  suite("until", function()
    test("until Exp do Exp", function()
      parser:parse("until true do print(x)")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        type = SentType.STMT,
        subtype = StmtType.UNTIL,
        decl = nil,
        value = nil
      })
      assert(stmt.cond:__tostring()):eq("true")
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(call print x)")
      assert(stmt.catch):isNil()
      assert(stmt.finally):isNil()

      assert(parser:next()):isNil()
    end)

    test("until Decl; Exp do Exp", function()
      parser:parse("until x := 1; x < 100 do print(x)")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        type = SentType.STMT,
        subtype = StmtType.UNTIL,
        decl = "var"
      })
      assert(stmt.value:__tostring()):eq("(= x 1)")
      assert(stmt.cond:__tostring()):eq("(< x 100)")
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(call print x)")
      assert(stmt.catch):isNil()
      assert(stmt.finally):isNil()

      assert(parser:next()):isNil()
    end)

    test("until Decl do Exp", function()
      parser:parse("until x := 1 do print(x)")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        type = SentType.STMT,
        subtype = StmtType.UNTIL,
        decl = "var"
      })
      assert(stmt.value:__tostring()):eq("(= x 1)")
      assert(stmt.cond):isNil()
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(call print x)")
      assert(stmt.catch):isNil()
      assert(stmt.finally):isNil()

      assert(parser:next()):isNil()
    end)

    test("until Exp do", function()
      parser:parse("until true do")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        type = SentType.STMT,
        subtype = StmtType.UNTIL,
        decl = nil,
        value = nil
      })
      assert(stmt.cond:__tostring()):eq("true")
      assert(stmt.body):isEmpty()
      assert(stmt.catch):isNil()
      assert(stmt.finally):isNil()

      assert(parser:next()):isNil()
    end)

    test("until Exp do\\n Sentence", function()
      parser:parse("until true do\n print(x)")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        type = SentType.STMT,
        subtype = StmtType.UNTIL,
        decl = nil,
        value = nil
      })
      assert(stmt.cond:__tostring()):eq("true")
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(call print x)")
      assert(stmt.catch):isNil()
      assert(stmt.finally):isNil()

      assert(parser:next()):isNil()
    end)

    test("until Decl; Exp do\\n Sentence", function()
      parser:parse("until x := 1; x < 100 do\n print(x)")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        type = SentType.STMT,
        subtype = StmtType.UNTIL,
        decl = "var"
      })
      assert(stmt.value:__tostring()):eq("(= x 1)")
      assert(stmt.cond:__tostring()):eq("(< x 100)")
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(call print x)")
      assert(stmt.catch):isNil()
      assert(stmt.finally):isNil()

      assert(parser:next()):isNil()
    end)

    test("until Exp do\\n Sentence\\n Sentence", function()
      parser:parse("until true do\n print(x)\n print(y)")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        type = SentType.STMT,
        subtype = StmtType.UNTIL,
        decl = nil,
        value = nil
      })
      assert(stmt.cond:__tostring()):eq("true")
      assert(stmt.body):len(2)
      assert(stmt.body[1]:__tostring()):eq("(call print x)")
      assert(stmt.body[2]:__tostring()):eq("(call print y)")
      assert(stmt.catch):isNil()
      assert(stmt.finally):isNil()

      assert(parser:next()):isNil()
    end)

    test("until Exp do\\n Sentence\\nSentence", function()
      parser:parse("until true do\n print(x)\nprint(y)")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        type = SentType.STMT,
        subtype = StmtType.UNTIL,
        decl = nil,
        value = nil
      })
      assert(stmt.cond:__tostring()):eq("true")
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(call print x)")
      assert(stmt.catch):isNil()
      assert(stmt.finally):isNil()

      assert(parser:next():__tostring()):eq("(call print y)")
      assert(parser:next()):isNil()
    end)

    test("until Exp do Block else Block", function()
      parser:parse("until true do\n print(x)\nelse\n print(y)")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        type = SentType.STMT,
        subtype = StmtType.UNTIL,
        decl = nil,
        value = nil
      })
      assert(stmt.cond:__tostring()):eq("true")
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(call print x)")
      assert(stmt.el):len(1)
      assert(stmt.el[1]:__tostring()):eq("(call print y)")
      assert(stmt.catch):isNil()
      assert(stmt.finally):isNil()

      assert(parser:next()):isNil()
    end)

    test("until Exp do\\n Sentence\\ncatch\\n  Sentence", function()
      parser:parse("until true do\n print(x)\ncatch\n  print(e)")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        type = SentType.STMT,
        subtype = StmtType.UNTIL,
        decl = nil,
        value = nil
      })
      assert(stmt.cond:__tostring()):eq("true")
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(call print x)")
      assert(stmt.catch):isTable()
      assert(stmt.catch.var):isNil()
      assert(stmt.catch.body):len(1)
      assert(stmt.catch.body[1]:__tostring()):eq("(call print e)")
      assert(stmt.finally):isNil()

      assert(parser:next()):isNil()
    end)

    test("until Exp do\\n print(x)\\ncatch\\nprint(e)", function()
      parser:parse("until true do\n print(x)\ncatch\nprint(e)")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        type = SentType.STMT,
        subtype = StmtType.UNTIL,
        decl = nil,
        value = nil
      })
      assert(stmt.cond:__tostring()):eq("true")
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(call print x)")
      assert(stmt.catch):isTable()
      assert(stmt.catch.var):isNil()
      assert(stmt.catch.body):len(0)
      assert(stmt.finally):isNil()

      assert(parser:next():__tostring()):eq("(call print e)")
      assert(parser:next()):isNil()
    end)

    test("until Exp do\\n print(x)\\ncatch e\\n  print(e)", function()
      parser:parse("until true do\n  print(x)\ncatch e\n  print(e)")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        type = SentType.STMT,
        subtype = StmtType.UNTIL,
        decl = nil,
        value = nil
      })
      assert(stmt.cond:__tostring()):eq("true")
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(call print x)")
      assert(stmt.catch):isTable()
      assert(stmt.catch.var):eq("e")
      assert(stmt.catch.body):len(1)
      assert(stmt.catch.body[1]:__tostring()):eq("(call print e)")
      assert(stmt.finally):isNil()

      assert(parser:next()):isNil()
    end)

    test("until Exp do\\n print(x)\\nfinally\\n  print(y)", function()
      parser:parse("until true do\n print(x)\nfinally\n print(y)")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        type = SentType.STMT,
        subtype = StmtType.UNTIL,
        decl = nil,
        value = nil
      })
      assert(stmt.cond:__tostring()):eq("true")
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(call print x)")
      assert(stmt.catch):isNil()
      assert(stmt.finally):isTable()
      assert(stmt.finally.body):len(1)
      assert(stmt.finally.body[1]:__tostring()):eq("(call print y)")

      assert(parser:next()):isNil()
    end)

    test("until Exp do\\n print(x)\\ncatch e\\n print(y)\\nfinally\\n print(z)", function()
      parser:parse("until true do\n print(x)\ncatch e\n print(y)\nfinally\n print(z)")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        type = SentType.STMT,
        subtype = StmtType.UNTIL,
        decl = nil,
        value = nil
      })
      assert(stmt.cond:__tostring()):eq("true")
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(call print x)")
      assert(stmt.catch):isTable()
      assert(stmt.catch.var):eq("e")
      assert(stmt.catch.body):len(1)
      assert(stmt.catch.body[1]:__tostring()):eq("(call print y)")
      assert(stmt.finally):isTable()
      assert(stmt.finally.body):len(1)
      assert(stmt.finally.body[1]:__tostring()):eq("(call print z)")

      assert(parser:next()):isNil()
    end)
  end):tags("until")

  --------
  -- do --
  --------
  suite("do", function()
    test("do\\n Sentence", function()
      parser:parse("do\n print(x)")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        type = SentType.STMT,
        subtype = StmtType.DO,
        cond = nil,
        catch = nil,
        finally = nil
      })
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(call print x)")
    end)

    test("do name := 1, name::= 2\\n Sentence", function()
      parser:parse("do x := 1+2, y ::= 2+1\n print(y)")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        type = SentType.STMT,
        subtype = StmtType.DO,
        cond = nil,
        catch = nil,
        finally = nil
      })

      assert(stmt.vars):len(2)
      assert(stmt.vars[1]):has({name = "x", kind = "var"})
      assert(tostring(stmt.vars[1].value)):eq("(+ 1 2)")
      assert(stmt.vars[2]):has({name = "y", kind = "const"})
      assert(tostring(stmt.vars[2].value)):eq("(+ 2 1)")

      assert(stmt.body):len(1)
      assert(tostring(stmt.body[1])):eq("(call print y)")
    end)

    test("error when var value not specified", function()
      parser:parse("do x, y ::= 2+1\n print(y)")
      assert(function() parser:next() end):raises("anonymous code (1, 5): ':=' or '::=' expected.")
    end)

    test("do\\n Sentence\\ncatch\\n Sentence", function()
      parser:parse("do\n print(x)\ncatch\n x += 1")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        type = SentType.STMT,
        subtype = StmtType.DO,
        cond = nil,
        finally = nil
      })
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(call print x)")
      assert(stmt.catch):isTable():has({
        var = nil
      })
      assert(stmt.catch.body):len(1)
      assert(stmt.catch.body[1]:__tostring()):eq("(+= x 1)")
    end)

    test("do\\n Sentence\\nfinally\\n Sentence", function()
      parser:parse("do\n print(x)\nfinally\n x += 1")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        type = SentType.STMT,
        subtype = StmtType.DO,
        cond = nil,
        catch = nil
      })
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(call print x)")
      assert(stmt.finally):isTable()
      assert(stmt.finally.body):len(1)
      assert(stmt.finally.body[1]:__tostring()):eq("(+= x 1)")
    end)

    test("do\\n Sentence\\ncatch\\n Sentence\\nfinally\\n Sentence", function()
      parser:parse("do\n print(x)\ncatch\n a += 1\nfinally\n b += 1")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        type = SentType.STMT,
        subtype = StmtType.DO,
        cond = nil,
      })
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(call print x)")
      assert(stmt.catch):isTable():has({
        var = nil
      })
      assert(stmt.catch.body):len(1)
      assert(stmt.catch.body[1]:__tostring()):eq("(+= a 1)")
      assert(stmt.finally):isTable()
      assert(stmt.finally.body):len(1)
      assert(stmt.finally.body[1]:__tostring()):eq("(+= b 1)")
    end)

    test("do\\n Sentence\\nwhile Exp", function()
      parser:parse("do\n print(x)\nwhile x == 1")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        type = SentType.STMT,
        subtype = StmtType.DO,
        catch = nil,
        finally = nil
      })
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(call print x)")
      assert(stmt.cond:__tostring()):eq("(== x 1)")
    end)

    test("do\\n Sentence\\nwhile Exp\\ncatch\\n Sentence\\nfinally\\n Sentence", function()
      parser:parse("do\n print(x)\nwhile x == 123\ncatch\n a += 1\nfinally\n b += 1")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        type = SentType.STMT,
        subtype = StmtType.DO,
      })
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(call print x)")
      assert(stmt.cond:__tostring()):eq("(== x 123)")
      assert(stmt.catch):isTable():has({
        var = nil
      })
      assert(stmt.catch.body):len(1)
      assert(stmt.catch.body[1]:__tostring()):eq("(+= a 1)")
      assert(stmt.finally):isTable()
      assert(stmt.finally.body):len(1)
      assert(stmt.finally.body[1]:__tostring()):eq("(+= b 1)")
    end)
  end)

  ---------
  -- for --
  ---------
  suite("for", function()
    test("for Name ; Exp do Exp", function()
      parser:parse("for i; i < 10 do print(i)")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        type = SentType.STMT,
        subtype = StmtType.FOR,
        def = {
          {name = "i", value = nil}
        },
        iter = nil,
        catch = nil,
        finally = nil
      })
      assert(tostring(stmt.cond)):eq("(< i 10)")
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(call print i)")
    end)

    test("for Name = Exp ; Exp do Exp", function()
      parser:parse("for i = 0; i < 10 do print(i)")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        type = SentType.STMT,
        subtype = StmtType.FOR,
        iter = nil,
        catch = nil,
        finally = nil
      })
      assert(stmt.def):len(1)
      assert(stmt.def[1].name):eq("i")
      assert(tostring(stmt.def[1].value)):eq("0")
      assert(tostring(stmt.cond)):eq("(< i 10)")
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(call print i)")
    end)

    test("for Name, Name ; Exp do Exp", function()
      parser:parse("for i, j; i < 10 do print(i)")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        type = SentType.STMT,
        subtype = StmtType.FOR,
        def = {
          {name = "i", value = nil},
          {name = "j", value = nil}
        },
        iter = nil,
        catch = nil,
        finally = nil
      })
      assert(tostring(stmt.cond)):eq("(< i 10)")
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(call print i)")
    end)

    test("for Name = Exp, Name = Exp ; Exp do Exp", function()
      parser:parse("for i = 0, j = 1; i < 10 do print(i)")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        type = SentType.STMT,
        subtype = StmtType.FOR,
        iter = nil,
        catch = nil,
        finally = nil
      })
      assert(stmt.def):len(2)
      assert(stmt.def[1].name):eq("i")
      assert(tostring(stmt.def[1].value)):eq("0")
      assert(stmt.def[2].name):eq("j")
      assert(tostring(stmt.def[2].value)):eq("1")
      assert(tostring(stmt.cond)):eq("(< i 10)")
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(call print i)")
    end)

    test("for Variable; Exp; Exp do Exp", function()
      parser:parse("for i = 0; i < 10; i += 1 do print(i)")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        type = SentType.STMT,
        subtype = StmtType.FOR,
        catch = nil,
        finally = nil
      })
      assert(stmt.def):len(1)
      assert(stmt.def[1].name):eq("i")
      assert(tostring(stmt.def[1].value)):eq("0")
      assert(tostring(stmt.cond)):eq("(< i 10)")
      assert(tostring(stmt.iter)):eq("(+= i 1)")
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(call print i)")
    end)

    test("for Variable; Exp; Exp do Block else Block", function()
      parser:parse('for i = 0; i < 10; i += 1 do print(i)\nelse print(j)')

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        type = SentType.STMT,
        subtype = StmtType.FOR,
        catch = nil,
        finally = nil
      })
      assert(stmt.def):len(1)
      assert(stmt.def[1].name):eq("i")
      assert(tostring(stmt.def[1].value)):eq("0")
      assert(tostring(stmt.cond)):eq("(< i 10)")
      assert(tostring(stmt.iter)):eq("(+= i 1)")
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(call print i)")
      assert(stmt.el):len(1)
      assert(stmt.el[1]:__tostring()):eq("(call print j)")
    end)
  end):tags("for")

  --------------
  -- for each --
  --------------
  suite("for each", function()
    test("error - :=, ::= or in expected", function()
      parser:parse("for each i list do print(i)")
      assert(function() parser:next() end):raises("anonymous code (1, 12): ':=', '::=' or 'in' expected.")
    end)

    test("await for each Name := Exp do Sentence", function()
      parser:parse("await for each i := list do print(i)")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        type = SentType.STMT,
        subtype = StmtType.FOR_EACH,
        await = true,
        decl = "var",
        key = nil,
        value = "i",
        valueType = nil,
        el = nil,
        catch = nil,
        finally = nil
      })
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(call print i)")
    end)

    test("for each Name in Exp do Sentence", function(params)
      parser:parse("for each i " .. params[1] .. " list do print(i)")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        type = SentType.STMT,
        subtype = StmtType.FOR_EACH,
        await = false,
        decl = "const",
        key = nil,
        value = "i",
        el = nil,
        catch = nil,
        finally = nil
      })
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(call print i)")
    end):iter(
      {subtitle=" # in", params="in"},
      {subtitle=" # ::=", params="::="}
    )

    test("for each Name: Type in Exp do Sentence", function(params)
      parser:parse("for each i: num " .. params[1] .. " list do print(i)")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        type = SentType.STMT,
        subtype = StmtType.FOR_EACH,
        await = false,
        decl = "const",
        key = nil,
        value = "i",
        valueType = "num",
        el = nil,
        catch = nil,
        finally = nil
      })
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(call print i)")
    end):iter(
      {subtitle=" # in", params="in"},
      {subtitle=" # ::=", params="::="}
    )

    test("for each Name ::= Exp do Sentence", function()
      parser:parse("for each i ::= list do print(i)")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        type = SentType.STMT,
        subtype = StmtType.FOR_EACH,
        await = false,
        decl = "const",
        key = nil,
        value = "i",
        el = nil,
        catch = nil,
        finally = nil
      })
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(call print i)")
    end)

    test("for each Name, Name in Exp do Sentence", function(params)
      parser:parse("for each i, j " .. params[1] .. " list do print(i, j)")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        type = SentType.STMT,
        subtype = StmtType.FOR_EACH,
        await = false,
        decl = "const",
        key = "i",
        value = "j",
        el = nil,
        catch = nil,
        finally = nil
      })
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(call print i j)")
    end):iter(
      {subtitle=" # in", params="in"},
      {subtitle=" # ::=", params="::="}
    )

    test("for each Name: Type, Name: Type in Exp do Sentence", function(params)
      parser:parse("for each i: num, j: text " .. params[1] .. " list do print(i, j)")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        type = SentType.STMT,
        subtype = StmtType.FOR_EACH,
        await = false,
        decl = "const",
        key = "i",
        keyType = "num",
        value = "j",
        valueType = "text",
        el = nil,
        catch = nil,
        finally = nil
      })
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(call print i j)")
    end):iter(
      {subtitle=" # in", params="in"},
      {subtitle=" # ::=", params="::="}
    )

    test("for each Name, Name := Exp do Sentence", function(params)
      parser:parse("for each i, j := list do print(i, j)")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        type = SentType.STMT,
        subtype = StmtType.FOR_EACH,
        await = false,
        decl = "var",
        key = "i",
        value = "j",
        el = nil,
        catch = nil,
        finally = nil
      })
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(call print i j)")
    end)

    test("for each Name in Exp do\\n Sentence", function()
      parser:parse("for each i in list do\n print(i)")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        type = SentType.STMT,
        subtype = StmtType.FOR_EACH,
        await = false,
        decl = "const",
        key = nil,
        value = "i",
        el = nil,
        catch = nil,
        finally = nil
      })
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(call print i)")
    end)

    test("for each Name, Name in Exp do\\n Sentence", function()
      parser:parse("for each i, j in list do\n print(i, j)")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        type = SentType.STMT,
        subtype = StmtType.FOR_EACH,
        await = false,
        decl = "const",
        key = "i",
        value = "j",
        el = nil,
        catch = nil,
        finally = nil
      })
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(call print i j)")
    end)

    test("for each Name, Name in Exp do Block else Block", function()
      parser:parse("for each i, j in list do\n print(i, j)\nelse print(x)")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        type = SentType.STMT,
        subtype = StmtType.FOR_EACH,
        await = false,
        decl = "const",
        key = "i",
        value = "j",
        catch = nil,
        finally = nil
      })
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(call print i j)")
      assert(stmt.el):len(1)
      assert(stmt.el[1]:__tostring()):eq("(call print x)")
    end)

    test("for each Name in Exp do\\n Sentence\\ncatch\\n Sentece\\nfinally\\n Sentence", function()
      parser:parse("for each i in list do\n print(i)\ncatch\n print(c)\nfinally\n print(f)")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        type = SentType.STMT,
        subtype = StmtType.FOR_EACH,
        await = false,
        decl = "const",
        key = nil,
        value = "i",
        el = nil
      })
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(call print i)")
      assert(stmt.catch):isTable():has({
        var = nil
      })
      assert(stmt.catch.body):len(1)
      assert(stmt.catch.body[1]:__tostring()):eq("(call print c)")
      assert(stmt.finally.body):len(1)
      assert(stmt.finally.body[1]:__tostring()):eq("(call print f)")
    end)
  end)

  ----------
  -- proc --
  ----------
  suite("proc", function()
    test("proc Name() = Exp - error, = is invalid", function()
      parser:parse("proc sum(x, y) = x + y")
      assert(function() parser:next() end):raises()
    end)

    test("pr Name.Name() Body", function()
      parser:parse("pr Calcul.sum(x, y) x + y")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        visib = nil,
        async = false,
        itype = "Calcul",
        name = "sum",
        accessor = nil,
        params = {
          {
            const = false,
            modifier = nil,
            name = "x",
            optional = false,
            type = nil,
            value = nil
          },
          {
            const = false,
            modifier = nil,
            name = "y",
            optional = false,
            type = nil,
            value = nil
          }
        }
      })
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(+ x y)")
    end)

    test("proc Name:Name() Body", function()
      parser:parse("proc Calcul:sum(x, y) x + y")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        visib = "intl",
        async = false,
        itype = "Calcul",
        name = "sum",
        accessor = nil,
        params = {
          {
            const = false,
            modifier = nil,
            name = "x",
            optional = false,
            type = nil,
            value = nil
          },
          {
            const = false,
            modifier = nil,
            name = "y",
            optional = false,
            type = nil,
            value = nil
          }
        }
      })
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(+ x y)")
    end)

    test("proc Name!Name() Body", function()
      parser:parse("proc Calcul!sum(x, y) x + y")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        visib = "pvt",
        async = false,
        itype = "Calcul",
        name = "sum",
        accessor = nil,
        params = {
          {
            const = false,
            modifier = nil,
            name = "x",
            optional = false,
            type = nil,
            value = nil
          },
          {
            const = false,
            modifier = nil,
            name = "y",
            optional = false,
            type = nil,
            value = nil
          }
        }
      })
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(+ x y)")
    end)

    test("proc Name() Body", function()
      parser:parse("proc sum(x, y) x + y")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        visib = nil,
        async = false,
        itype = nil,
        name = "sum",
        accessor = nil,
        params = {
          {
            const = false,
            modifier = nil,
            name = "x",
            optional = false,
            type = nil,
            value = nil
          },
          {
            const = false,
            modifier = nil,
            name = "y",
            optional = false,
            type = nil,
            value = nil
          }
        }
      })
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(+ x y)")
    end)

    test("export proc Name() Body", function()
      parser:parse("export proc sum(x, y) x + y")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        visib = "export",
        async = false,
        itype = nil,
        name = "sum",
        accessor = nil,
        params = {
          {
            const = false,
            modifier = nil,
            name = "x",
            optional = false,
            type = nil,
            value = nil
          },
          {
            const = false,
            modifier = nil,
            name = "y",
            optional = false,
            type = nil,
            value = nil
          }
        }
      })
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(+ x y)")
    end)

    test("export proc() Body", function()
      parser:parse("export proc(x, y) x + y")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        visib = "export",
        async = false,
        itype = nil,
        name = "__anonymous__",
        accessor = nil,
        params = {
          {
            const = false,
            modifier = nil,
            name = "x",
            optional = false,
            type = nil,
            value = nil
          },
          {
            const = false,
            modifier = nil,
            name = "y",
            optional = false,
            type = nil,
            value = nil
          }
        }
      })
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(+ x y)")
    end)

    test("error when proc name expected", function()
      parser:parse("pub proc(x, y) x + y")
      assert(function() parser:next() end):raises("anonymous code (1, 9): name expected.")
    end)

    test("async proc Name() Body", function()
      parser:parse("async proc sum(x, y) x + y")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        visib = nil,
        async = true,
        itype = nil,
        name = "sum",
        accessor = nil,
        params = {
          {
            const = false,
            modifier = nil,
            name = "x",
            optional = false,
            type = nil,
            value = nil
          },
          {
            const = false,
            modifier = nil,
            name = "y",
            optional = false,
            type = nil,
            value = nil
          }
        }
      })
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(+ x y)")
    end)

    test("pub async proc Name() Body", function()
      parser:parse("pub async proc sum(x, y) x + y")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        visib = "pub",
        async = true,
        itype = nil,
        name = "sum",
        accessor = nil,
        params = {
          {
            const = false,
            modifier = nil,
            name = "x",
            optional = false,
            type = nil,
            value = nil
          },
          {
            const = false,
            modifier = nil,
            name = "y",
            optional = false,
            type = nil,
            value = nil
          }
        }
      })
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(+ x y)")
    end)
  end)

  --------
  -- fn --
  --------
  suite("fn", function()
    test("fn with subparams", function()
      --(1) act
      parser:parse("fn fun(a)(b)(c) = a + b + c")
      stmt = parser:next()

      --(2) assert
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        visib = nil,
        async = false,
        itype = nil,
        name = "fun",
        accessor = nil,
        params = {
          {
            const = false,
            modifier = nil,
            name = "a",
            optional = false,
            type = nil,
            value = nil
          }
        },
        subparams = {
          {
            {
              const = false,
              modifier = nil,
              name = "b",
              optional = false,
              type = nil,
              value = nil
            }
          },
          {
            {
              const = false,
              modifier = nil,
              name = "c",
              optional = false,
              type = nil,
              value = nil
            }
          }
        }
      })

      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("return (+ (+ a b) c)")
    end)

    test("export fn() Body", function()
      parser:parse("export fn(x, y) = x + y")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        visib = "export",
        async = false,
        itype = nil,
        name = "__anonymous__",
        accessor = nil,
        params = {
          {
            const = false,
            modifier = nil,
            name = "x",
            optional = false,
            type = nil,
            value = nil
          },
          {
            const = false,
            modifier = nil,
            name = "y",
            optional = false,
            type = nil,
            value = nil
          }
        }
      })
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("return (+ x y)")
    end)

    test("error when fn name expected", function()
      parser:parse("pub fn(x, y) = x + y")
      assert(function() parser:next() end):raises("anonymous code (1, 7): name expected.")
    end)

    suite("prop", function()
      test("fn name = exp", function()
        parser:parse("fn x = y")
        stmt = parser:next()
        assert(stmt):isTable():has({
          line = 1,
          col = 1,
          visib = nil,
          itype = nil,
          name = "x",
          accessor = nil,
          rtype = nil,
          rvar = nil,
          params = {},
        })
        assert(stmt.annots:has("prop")):eq(true)
        assert(stmt.body):len(1)
        assert(stmt.body[1]:__tostring()):eq("return y")
      end)
    end)

    suite("params", function()
      test("type or value expected", function()
        parser:parse("fn f(params: ^")
        assert(function() parser:next() end):raises("anonymous code (1, 14): Type or value expected.")
      end)

      suite("=>", function()
        test("fn Name(Name=>[Name])", function()
          parser:parse("fn f(params=>[x])")

          stmt = parser:next()
          assert(stmt):isTable():has({
            line = 1,
            col = 1,
            visib = nil,
            itype = nil,
            name = "f",
            accessor = nil,
            rtype = nil,
            rvar = nil,
            params = {
              {
                const = false,
                optional = false,
                name = "params",
                type = {kind="be", type={kind="type", name="list"}},
                uVars = {"x"}
              }
            }
          })
          assert(stmt.body):isEmpty()
        end)

        test("fn Name(Name=>{Name})", function()
          parser:parse("fn f(params=>{x})")

          stmt = parser:next()
          assert(stmt):isTable():has({
            line = 1,
            col = 1,
            visib = nil,
            itype = nil,
            name = "f",
            accessor = nil,
            rtype = nil,
            rvar = nil,
            params = {
              {
                const = false,
                optional = false,
                name = "params",
                type = {kind="be", type={kind="type", name="map"}},
                uVars = {"x"}
              }
            }
          })
          assert(stmt.body):isEmpty()
        end)

        test("fn Name(Name=>[Name,Name])", function()
          parser:parse("fn f(params=>[x,y])")

          stmt = parser:next()
          assert(stmt):isTable():has({
            line = 1,
            col = 1,
            visib = nil,
            itype = nil,
            name = "f",
            accessor = nil,
            rtype = nil,
            rvar = nil,
            params = {
              {
                const = false,
                optional = false,
                name = "params",
                type = {kind="be", type={kind="type", name="list"}},
                uVars = {"x", "y"}
              }
            }
          })
          assert(stmt.body):isEmpty()
        end)

        test("fn Name(Name=>{Name,Name})", function()
          parser:parse("fn f(params=>{x,y})")

          stmt = parser:next()
          assert(stmt):isTable():has({
            line = 1,
            col = 1,
            visib = nil,
            itype = nil,
            name = "f",
            accessor = nil,
            rtype = nil,
            rvar = nil,
            params = {
              {
                const = false,
                optional = false,
                name = "params",
                type = {kind="be", type={kind="type", name="map"}},
                uVars = {"x", "y"}
              }
            }
          })
          assert(stmt.body):isEmpty()
        end)

        test("fn Name(Name:{field:type}=>{...})", function()
          parser:parse("fn f(params:{x:num}=>{x,y})")

          stmt = parser:next()
          assert(stmt):isTable():has({
            line = 1,
            col = 1,
            visib = nil,
            itype = nil,
            name = "f",
            accessor = nil,
            rtype = nil,
            rvar = nil,
            params = {
              {
                const = false,
                optional = false,
                name = "params",
                type = {
                  kind = "map",
                  fields = {
                    {
                      name = "x",
                      type = {kind="be", type={kind="type", name="num"}},
                      mandatory = true
                    }
                  }
                },
                uVars = {"x", "y"}
              }
            }
          })
          assert(stmt.body):isEmpty()
        end)

        test("fn Name(Name=>) - error", function()
          parser:parse("fn f(params=>x)")
          assert(function() parser:next() end):raises("anonymous code (1, 14):")
        end)
      end)

      suite(":>", function()
        test("fn Name(Name:>[Name]) - { expected", function()
          parser:parse("fn f(params:>[x])")
          assert(function() parser:next() end):raises("anonymous code (1, 14): '{' expected.")
        end)

        test("fn Name(Name:>{Name, Name: *Type})", function()
          parser:parse("fn f(params:>{x, y: *num})")

          stmt = parser:next()
          assert(stmt):isTable():has({
            line = 1,
            col = 1,
            visib = nil,
            itype = nil,
            name = "f",
            accessor = nil,
            rtype = nil,
            rvar = nil,
            params = {
              {
                const = false,
                optional = false,
                name = "params",
                type = {
                  kind = "map",
                  fields = {
                    {name = "x", mandatory = true, type = nil},
                    {name = "y", mandatory = true, byCopy = true, type = {type = {kind = "type", name = "num"}, kind = "be"}}
                  }
                },
                uVars = {"x", "y"}
              }
            }
          })
          assert(stmt.body):isEmpty()
        end)

        test("fn Name(Name:>{...})", function()
          parser:parse("fn f(params:>{a,b:num,c?:text,d:list(num)})")

          stmt = parser:next()
          assert(stmt):isTable():has({
            line = 1,
            col = 1,
            visib = nil,
            itype = nil,
            name = "f",
            accessor = nil,
            rtype = nil,
            rvar = nil,
            params = {
              {
                const = false,
                optional = false,
                name = "params",
                type = {
                  kind = "map",
                  fields = {
                    {name="a", type=nil, mandatory=true},
                    {name="b", type={kind="be", type={kind="type", name="num"}}, mandatory=true},
                    {name="c", type={kind="be", type={kind="type", name="text"}}, mandatory=false},
                    {
                      name = "d",
                      type = {
                        kind = "oneOf",
                        types = {
                          {kind="be", type={kind="type", name="list"}},
                          {kind="be", type={kind="type", name="num"}}
                        },
                        call = {kind="exclusive",object="list"},
                      },
                      mandatory=true
                    }
                  }
                },
                uVars = {"a", "b", "c", "d"}
              }
            }
          })
          assert(stmt.body):isEmpty()
        end)
      end)

      test("fn Name ( Name )", function()
        parser:parse("fn inc(x)")

        stmt = parser:next()
        assert(stmt):isTable():has({
          line = 1,
          col = 1,
          visib = nil,
          itype = nil,
          name = "inc",
          accessor = nil,
          rtype = nil,
          rvar = nil,
          params = {
            {
              const = false,
              modifier = nil,
              name = "x",
              optional = false,
              type = nil,
              value = nil,
              uVars = nil
            }
          }
        })
        assert(stmt.body):isEmpty()
      end)

      test("fn Name ( const Name )", function()
        parser:parse("fn inc(const x)")

        stmt = parser:next()
        assert(stmt):isTable():has({
          line = 1,
          col = 1,
          visib = nil,
          itype = nil,
          name = "inc",
          accessor = nil,
          rtype = nil,
          rvar = nil,
          params = {
            {
              const = true,
              modifier = nil,
              name = "x",
              optional = false,
              type = nil,
              value = nil,
              uVars = nil
            }
          }
        })
        assert(stmt.body):isEmpty()
      end)

      test("fn Name ( Name ? )", function()
        parser:parse("fn inc(x?)")

        stmt = parser:next()
        assert(stmt):isTable():has({
          line = 1,
          col = 1,
          visib = nil,
          itype = nil,
          name = "inc",
          accessor = nil,
          rtype = nil,
          rvar = nil,
          params = {
            {
              const = false,
              modifier = nil,
              name = "x",
              optional = true,
              type = nil,
              value = nil
            }
          }
        })
        assert(stmt.body):isEmpty()
      end)

      test("fn Name ( Name: Name, Name: *Name )", function()
        parser:parse("fn inc(x: num, y: *map)")

        stmt = parser:next()
        assert(stmt):isTable():has({
          line = 1,
          col = 1,
          visib = nil,
          itype = nil,
          name = "inc",
          accessor = nil,
          rtype = nil,
          rvar = nil,
          params = {
            {
              const = false,
              modifier = nil,
              name = "x",
              optional = false,
              type = {kind="be", type={kind="type", name="num"}},
              value = nil
            },

            {
              const = false,
              modifier = nil,
              name = "y",
              optional = false,
              byCopy = true,
              type = {kind="be", type={kind="type", name="map"}},
              value = nil
            }
          }
        })
        assert(stmt.body):isEmpty()
      end)

      test("fn Name ( Name: nil )", function()
        parser:parse("fn inc(x:nil)")

        stmt = parser:next()
        assert(stmt):isTable():has({
          line = 1,
          col = 1,
          visib = nil,
          itype = nil,
          name = "inc",
          accessor = nil,
          rtype = nil,
          rvar = nil,
          params = {
            {
              const = false,
              modifier = nil,
              name = "x",
              optional = false,
              type = {kind="be", type={kind="value", value=nil}},
              value = nil
            }
          }
        })
        assert(stmt.body):isEmpty()
      end)

      test('fn Name ( Name : "Text" )', function()
        parser:parse('fn inc(x:"num")')

        stmt = parser:next()
        assert(stmt):isTable():has({
          line = 1,
          col = 1,
          visib = nil,
          itype = nil,
          name = "inc",
          accessor = nil,
          rtype = nil,
          rvar = nil,
          params = {
            {
              const = false,
              modifier = nil,
              name = "x",
              optional = false,
              type = {kind="be", type={kind="value", value="num"}},
              value = nil
            }
          }
        })
        assert(stmt.body):isEmpty()
      end)

      test("fn Name ( Name : Num )", function()
        parser:parse('fn inc(x:123)')

        stmt = parser:next()
        assert(stmt):isTable():has({
          line = 1,
          col = 1,
          visib = nil,
          itype = nil,
          name = "inc",
          accessor = nil,
          rtype = nil,
          rvar = nil,
          params = {
            {
              const = false,
              modifier = nil,
              name = "x",
              optional = false,
              type = {kind="be", type={kind="value", value=123}},
              value = nil
            }
          }
        })
        assert(stmt.body):isEmpty()
      end)

      test("fn Name ( Name : Bool )", function(params)
        parser:parse(string.format('fn inc(x:%s)', params[1]))

        stmt = parser:next()
        assert(stmt):isTable():has({
          line = 1,
          col = 1,
          visib = nil,
          itype = nil,
          name = "inc",
          accessor = nil,
          rtype = nil,
          rvar = nil,
          params = {
            {
              const = false,
              modifier = nil,
              name = "x",
              optional = false,
              type = {kind="be", type={kind="value", value=params[1]}},
              value = nil
            }
          }
        })
        assert(stmt.body):isEmpty()
      end):iter(
        {subtitle = " # true", params = {true}},
        {subtitle = " # false", params = {false}}
      )

      test("fn Name ( Name ?: Name )", function()
        parser:parse("fn inc(x?:num)")

        stmt = parser:next()
        assert(stmt):isTable():has({
          line = 1,
          col = 1,
          visib = nil,
          itype = nil,
          name = "inc",
          accessor = nil,
          rtype = nil,
          rvar = nil,
          params = {
            {
              const = false,
              modifier = nil,
              name = "x",
              optional = true,
              type = {kind="be", type={kind="type", name="num"}},
              value = nil
            }
          }
        })
        assert(stmt.body):isEmpty()
      end)

      test("fn Name ( Name ?: Name )", function()
        parser:parse('fn inc(x?:"num")')

        stmt = parser:next()
        assert(stmt):isTable():has({
          line = 1,
          col = 1,
          visib = nil,
          itype = nil,
          name = "inc",
          accessor = nil,
          rtype = nil,
          rvar = nil,
          params = {
            {
              const = false,
              modifier = nil,
              name = "x",
              optional = true,
              type = {kind="be", type={kind="value", value="num"}},
              value = nil
            }
          }
        })
        assert(stmt.body):isEmpty()
      end)

      test("fn Name ( pub Name )", function(params)
        parser:parse(string.format("fn inc(%s x)", params[1]))

        stmt = parser:next()
        assert(stmt):isTable():has({
          line = 1,
          col = 1,
          visib = nil,
          itype = nil,
          name = "inc",
          accessor = nil,
          rtype = nil,
          rvar = nil,
          params = {
            {
              const = false,
              modifier = ".",
              name = "x",
              optional = false,
              type = nil,
              value = nil
            }
          }
        })
        assert(stmt.body):isEmpty()
      end):iter(
        {subtitle = " # pub", params = "pub"},
        {subtitle = " # .", params = "."}
      )

      test("fn Name ( intl Name )", function(params)
        parser:parse(string.format("fn inc(%s x)", params[1]))

        stmt = parser:next()
        assert(stmt):isTable():has({
          line = 1,
          col = 1,
          visib = nil,
          itype = nil,
          name = "inc",
          accessor = nil,
          rtype = nil,
          rvar = nil,
          params = {
            {
              const = false,
              modifier = ":",
              name = "x",
              optional = false,
              type = nil,
              value = nil
            }
          }
        })
        assert(stmt.body):isEmpty()
      end):iter(
        {subtitle = " # intl", params = "intl"},
        {subtitle = " # :", params = ":"}
      )

      test("fn Name ( pvt Name )", function(params)
        parser:parse(string.format("fn inc(%s x)", params[1]))

        stmt = parser:next()
        assert(stmt):isTable():has({
          line = 1,
          col = 1,
          visib = nil,
          itype = nil,
          name = "inc",
          accessor = nil,
          rtype = nil,
          rvar = nil,
          params = {
            {
              const = false,
              modifier = "!",
              name = "x",
              optional = false,
              type = nil,
              value = nil
            }
          }
        })
        assert(stmt.body):isEmpty()
      end):iter(
        {subtitle = " # pvt", params = "pvt"},
        {subtitle = " # !", params = "!"}
      )

      test("fn Name ( ... Name )", function()
        parser:parse("fn inc(...x)")

        stmt = parser:next()
        assert(stmt):isTable():has({
          line = 1,
          col = 1,
          visib = nil,
          itype = nil,
          name = "inc",
          rtype = nil,
          rvar = nil,
          accessor = nil,
          params = {
            {
              const = false,
              modifier = "...",
              name = "x",
              optional = false,
              type = nil,
              value = nil
            }
          }
        })
        assert(stmt.body):isEmpty()
      end)

      test("fn Name(Name : {})", function()
        parser:parse("fn myfn(p:{})")
        stmt = parser:next()
        assert(stmt):isTable():has({
          line = 1,
          col = 1,
          visib = nil,
          itype = nil,
          name = "myfn",
          accessor = nil,
          rtype = nil,
          rvar = nil,
          body = {}
        })
        assert(stmt.params):len(1)
        assert(stmt.params[1]):has({
          const = false,
          modifier = nil,
          name = "p",
          optional = false,
          type = {kind="map", fields={}}
        })
      end)

      test("fn Name(Name : {Name})", function()
        parser:parse("fn myfn(p:{x})")
        stmt = parser:next()
        assert(stmt):isTable():has({
          line = 1,
          col = 1,
          visib = nil,
          itype = nil,
          name = "myfn",
          accessor = nil,
          rtype = nil,
          rvar = nil,
          body = {}
        })
        assert(stmt.params):len(1)
        assert(stmt.params[1]):has({
          const = false,
          modifier = nil,
          name = "p",
          optional = false,
          type = {
            kind = "map",
            fields = {
              {name="x", type=nil, mandatory=true}
            }
          }
        })
      end)

      test("fn Name(Name : {Name?})", function()
        parser:parse("fn myfn(p:{x?})")
        stmt = parser:next()
        assert(stmt):isTable():has({
          line = 1,
          col = 1,
          visib = nil,
          itype = nil,
          name = "myfn",
          accessor = nil,
          rtype = nil,
          rvar = nil,
          body = {}
        })
        assert(stmt.params):len(1)
        assert(stmt.params[1]):has({
          const = false,
          modifier = nil,
          name = "p",
          optional = false,
          type = {
            kind = "map",
            fields = {
              {name="x", type=nil, mandatory=false}
            }
          }
        })
      end)

      test("fn Name(Name : {Name : Name})", function()
        parser:parse("fn myfn(p:{x:num})")
        stmt = parser:next()
        assert(stmt):isTable():has({
          line = 1,
          col = 1,
          visib = nil,
          itype = nil,
          name = "myfn",
          accessor = nil,
          rtype = nil,
          rvar = nil,
          body = {}
        })
        assert(stmt.params):len(1)
        assert(stmt.params[1]):has({
          const = false,
          modifier = nil,
          name = "p",
          optional = false,
          type = {
            kind = "map",
            fields = {
              {name="x", type={kind="be", type={kind="type", name="num"}}, mandatory=true}
            }
          }
        })
      end)

      test("fn Name(Name : {Name ?: Name})", function()
        parser:parse("fn myfn(p:{x?:num})")
        stmt = parser:next()
        assert(stmt):isTable():has({
          line = 1,
          col = 1,
          visib = nil,
          itype = nil,
          name = "myfn",
          accessor = nil,
          rtype = nil,
          rvar = nil,
          body = {}
        })
        assert(stmt.params):len(1)
        assert(stmt.params[1]):has({
          const = false,
          modifier = nil,
          name = "p",
          optional = false,
          type = {
            kind = "map",
            fields = {
              {name="x", type={kind="be", type={kind="type", name="num"}}, mandatory=false}
            }
          }
        })
      end)

      test("fn Name(Name : {Name : (Name)})", function()
        parser:parse("fn myfn(p:{x:(text)})")
        stmt = parser:next()
        assert(stmt):isTable():has({
          line = 1,
          col = 1,
          visib = nil,
          itype = nil,
          name = "myfn",
          accessor = nil,
          rtype = nil,
          rvar = nil,
          body = {}
        })
        assert(stmt.params):len(1)
        assert(stmt.params[1]):has({
          const = false,
          modifier = nil,
          name = "p",
          optional = false,
          type = {
            kind = "map",
            fields = {
              {
                name = "x",
                type = {
                  kind = "oneOf",
                  types = {{kind="be", type={kind="type", name="text"}}}
                },
                mandatory = true
              }
            }
          }
        })
      end)

      test("fn Name(Name : {Name : (Name, Text)})", function()
        parser:parse('fn myfn(p:{x:(text,"list")})')
        stmt = parser:next()
        assert(stmt):isTable():has({
          line = 1,
          col = 1,
          visib = nil,
          itype = nil,
          name = "myfn",
          accessor = nil,
          rtype = nil,
          rvar = nil,
          body = {}
        })
        assert(stmt.params):len(1)
        assert(stmt.params[1]):has({
          const = false,
          modifier = nil,
          name = "p",
          optional = false,
          type = {
            kind = "map",
            fields = {
              {
                name = "x",
                type = {
                  kind = "oneOf",
                  types = {
                    {kind="be", type={kind="type", name="text"}},
                    {kind="be", type={kind="value", value="list"}}
                  }
                },
                mandatory = true
              }
            }
          }
        })
      end)

      test("fn Name(Name : {Name : Name(Name)})", function()
        parser:parse("fn myfn(p:{x: list(text)})")
        stmt = parser:next()
        assert(stmt):isTable():has({
          line = 1,
          col = 1,
          visib = nil,
          itype = nil,
          name = "myfn",
          accessor = nil,
          rtype = nil,
          rvar = nil,
          body = {}
        })
        assert(stmt.params):len(1)
        assert(stmt.params[1]):has({
          const = false,
          modifier = nil,
          name = "p",
          optional = false,
          type = {
            kind = "map",
            fields = {
              {
                name = "x",
                type = {
                  kind = "oneOf",
                  types = {
                    {kind="be", type={kind="type", name="list"}},
                    {kind="be", type={kind="type", name="text"}}
                  },
                  call = {kind="exclusive", object="list"},
                },
                mandatory = true
              }
            }
          }
        })
      end)

      test("fn Name(Name : {Name : Name(Name, Text)})", function()
        parser:parse('fn myfn(p:{x: list(text,"num")})')
        stmt = parser:next()
        assert(stmt):isTable():has({
          line = 1,
          col = 1,
          visib = nil,
          itype = nil,
          name = "myfn",
          accessor = nil,
          rtype = nil,
          rvar = nil,
          body = {}
        })
        assert(stmt.params):len(1)
        assert(stmt.params[1]):has({
          const = false,
          modifier = nil,
          name = "p",
          optional = false,
          type = {
            kind = "map",
            fields = {
              {
                name = "x",
                type = {
                  kind = "oneOf",
                  types = {
                    {kind="be", type={kind="type", name="list"}},
                    {kind="be", type={kind="type", name="text"}},
                    {kind="be", type={kind="value", value="num"}}
                  },
                  call = {kind="exclusive", object="list"}
                },
                mandatory = true
              }
            }
          }
        })
      end)

      test("fn Name(Name : {Name, Name?:Name})", function()
        parser:parse("fn myfn(p:{x, y?:text})")
        stmt = parser:next()
        assert(stmt):isTable():has({
          line = 1,
          col = 1,
          visib = nil,
          itype = nil,
          name = "myfn",
          accessor = nil,
          rtype = nil,
          rvar = nil,
          body = {}
        })
        assert(stmt.params):len(1)
        assert(stmt.params[1]):has({
          const = false,
          modifier = nil,
          name = "p",
          optional = false,
          type = {
            kind = "map",
            fields = {
              {name="x", type=nil, mandatory=true},
              {name="y", type={kind="be", type={kind="type", name="text"}}, mandatory=false}
            }
          }
        })
      end)

      suite("multiple types", function()
        test("fn Name (Name : (Name))", function()
          parser:parse("fn myfn(p: (text))")
          stmt = parser:next()
          assert(stmt):isTable():has({
            line = 1,
            col = 1,
            visib = nil,
            itype = nil,
            name = "myfn",
            accessor = nil,
            rtype = nil,
            rvar = nil,
            body = {},
            params = {
              {
                const = false,
                modifier = nil,
                name = "p",
                optional = false,
                type = {kind="oneOf", types={{kind="be", type={kind="type", name="text"}}}}
              }
            }
          })
        end)

        test("fn Name (Name : (Name, Text))", function()
          parser:parse('fn myfn(p: (text,"list"))')
          stmt = parser:next()
          assert(stmt):isTable():has({
            line = 1,
            col = 1,
            visib = nil,
            itype = nil,
            name = "myfn",
            accessor = nil,
            rtype = nil,
            rvar = nil,
            body = {},
            params = {
              {
                const = false,
                modifier = nil,
                name = "p",
                optional = false,
                type = {
                  kind = "oneOf",
                  types = {
                    {kind="be", type={kind="type", name="text"}},
                    {kind="be", type={kind="value", value="list"}}
                  }
                }
              }
            }
          })
        end)

        test("fn Name (Name : (Name, Text, Num, Bool))", function()
          parser:parse('fn myfn(p?: (text,"list", 123, true, false))')
          stmt = parser:next()
          assert(stmt):isTable():has({
            line = 1,
            col = 1,
            visib = nil,
            itype = nil,
            name = "myfn",
            accessor = nil,
            rtype = nil,
            rvar = nil,
            body = {},
            params = {
              {
                const = false,
                modifier = nil,
                name = "p",
                optional = true,
                type = {
                  kind = "oneOf",
                  types = {
                    {kind="be", type={kind="type", name="text"}},
                    {kind="be", type={kind="value", value="list"}},
                    {kind="be", type={kind="value", value=123}},
                    {kind="be", type={kind="value", value=true}},
                    {kind="be", type={kind="value", value=false}}
                  }
                }
              }
            }
          })
        end)

        test("fn Name (Name : Name(Name))", function()
          parser:parse("fn myfn(p: list(text))")
          stmt = parser:next()
          assert(stmt):isTable():has({
            line = 1,
            col = 1,
            visib = nil,
            itype = nil,
            name = "myfn",
            accessor = nil,
            rtype = nil,
            rvar = nil,
            body = {},
            params = {
              {
                const = false,
                modifier = nil,
                name = "p",
                optional = false,
                type = {
                  kind = "oneOf",
                  types = {
                    {kind="be", type={kind="type", name="list"}},
                    {kind="be", type={kind="type", name="text"}}
                  },
                  call = {kind="exclusive", object="list"}
                }
              }
            }
          })
        end)

        test("fn Name (Name : Name(Name, Text))", function()
          parser:parse('fn myfn(p: list(text,"num"))')
          stmt = parser:next()
          assert(stmt):isTable():has({
            line = 1,
            col = 1,
            visib = nil,
            itype = nil,
            name = "myfn",
            accessor = nil,
            rtype = nil,
            rvar = nil,
            body = {},
            params = {
              {
                const = false,
                modifier = nil,
                name = "p",
                optional = false,
                type = {
                  kind = "oneOf",
                  types = {
                    {kind="be", type={kind="type", name="list"}},
                    {kind="be", type={kind="type", name="text"}},
                    {kind="be", type={kind="value", value="num"}},
                  },
                  call = {kind="exclusive", object="list"}
                }
              }
            }
          })
        end)

        test("fn Name (Name?: Name(Name, Text, Num, Bool))", function()
          parser:parse('fn myfn(p?: list(text,"num",123,true,false))')
          stmt = parser:next()
          assert(stmt):isTable():has({
            line = 1,
            col = 1,
            visib = nil,
            itype = nil,
            name = "myfn",
            accessor = nil,
            rtype = nil,
            rvar = nil,
            body = {},
            params = {
              {
                const = false,
                modifier = nil,
                name = "p",
                optional = true,
                type = {
                  kind = "oneOf",
                  types = {
                    {kind="be", type={kind="type", name="list"}},
                    {kind="be", type={kind="type", name="text"}},
                    {kind="be", type={kind="value", value="num"}},
                    {kind="be", type={kind="value", value=123}},
                    {kind="be", type={kind="value", value=true}},
                    {kind="be", type={kind="value", value=false}}
                  },
                  call = {kind="exclusive", object="list"}
                }
              }
            }
          })
        end)
      end)

      test("fn Name ( Name := LiteralNum )", function()
        parser:parse("fn inc(x := 123)")

        stmt = parser:next()
        assert(stmt):isTable():has({
          line = 1,
          col = 1,
          visib = nil,
          itype = nil,
          name = "inc",
          accessor = nil,
          rtype = nil,
          rvar = nil
        })
        assert(stmt.params):len(1)
        assert(stmt.params[1]):has({
          const = false,
          modifier = nil,
          name = "x",
          optional = true,
          type = {kind="be", type={kind="type", name="num"}},
        })
        assert(stmt.params[1].value:__tostring()):eq("123")
        assert(stmt.body):isEmpty()
      end)

      test("fn Name ( Name := LiteralText )", function()
        parser:parse('fn inc(x := "texto")')

        stmt = parser:next()
        assert(stmt):isTable():has({
          line = 1,
          col = 1,
          visib = nil,
          itype = nil,
          name = "inc",
          accessor = nil,
          rtype = nil,
          rvar = nil
        })
        assert(stmt.params):len(1)
        assert(stmt.params[1]):has({
          const = false,
          modifier = nil,
          name = "x",
          optional = true,
          type = {kind="be", type={kind="type", name="text"}},
        })
        assert(stmt.params[1].value:__tostring()):eq("texto")
        assert(stmt.body):isEmpty()
      end)

      test("fn Name ( Name := true )", function()
        parser:parse("fn inc(x := true)")

        stmt = parser:next()
        assert(stmt):isTable():has({
          line = 1,
          col = 1,
          visib = nil,
          itype = nil,
          name = "inc",
          accessor = nil,
          rtype = nil,
          rvar = nil
        })
        assert(stmt.params):len(1)
        assert(stmt.params[1]):has({
          const = false,
          modifier = nil,
          name = "x",
          optional = true,
          type = {kind="be", type={kind="type", name="bool"}},
        })
        assert(stmt.params[1].value:__tostring()):eq("true")
        assert(stmt.body):isEmpty()
      end)

      test("fn Name ( Name := false )", function()
        parser:parse("fn inc(x := false)")

        stmt = parser:next()
        assert(stmt):isTable():has({
          line = 1,
          col = 1,
          visib = nil,
          itype = nil,
          name = "inc",
          accessor = nil,
          rtype = nil,
          rvar = nil
        })
        assert(stmt.params):len(1)
        assert(stmt.params[1]):has({
          const = false,
          modifier = nil,
          name = "x",
          optional = true,
          type = {kind="be", type={kind="type", name="bool"}}
        })
        assert(stmt.params[1].value:__tostring()):eq("false")
        assert(stmt.body):isEmpty()
      end)

      test("fn Name ( Name := list )", function()
        parser:parse("fn inc(x := [])")

        stmt = parser:next()
        assert(stmt):isTable():has({
          line = 1,
          col = 1,
          visib = nil,
          itype = nil,
          name = "inc",
          accessor = nil,
          rtype = nil,
          rvar = nil
        })
        assert(stmt.params):len(1)
        assert(stmt.params[1]):has({
          const = false,
          modifier = nil,
          name = "x",
          optional = true,
          type = {kind="be", type={kind="type", name="list"}},
        })
        assert(tostring(stmt.params[1].value)):eq("[]")
        assert(stmt.body):isEmpty()
      end)

      test("fn Name ( Name := map )", function()
        parser:parse("fn inc(x := {})")

        stmt = parser:next()
        assert(stmt):isTable():has({
          line = 1,
          col = 1,
          visib = nil,
          itype = nil,
          name = "inc",
          accessor = nil,
          rtype = nil,
          rvar = nil
        })
        assert(stmt.params):len(1)
        assert(stmt.params[1]):has({
          const = false,
          modifier = nil,
          name = "x",
          optional = true,
          type = {kind="be", type={kind="type", name="map"}}
        })
        assert(tostring(stmt.params[1].value)):eq("{}")
        assert(stmt.body):isEmpty()
      end)

      test("fn Name ( Name := nil ) - error", function()
        parser:parse("fn inc(x := nil)")
        assert(function() parser:next() end):raises("anonymous code (1, 13): for infering type, the default value must be a literal: bool, list, map, num or text.")
      end)

      test("fn Name . Name () : 123", function()
        parser:parse("fn MyType.method() : 123")
        assert(function() parser:next() end):raises("anonymous code (1, 22): name expected.")
      end)
    end):tags("parameters")

    test("export async fn Name() Body", function()
      parser:parse("export async fn sum() return x + y")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        visib = "export",
        async = true,
        itype = nil,
        name = "sum",
        accessor = nil,
        params = {}
      })
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("return (+ x y)")
    end)

    test("fn Name ( Name ? = Exp ) = Exp - error", function()
      parser:parse("fn inc(x ? = 123) = x + 1")
      assert(function() parser:next() end):raises("anonymous code (1, 12): ')' expected.")
    end)

    test("fn Name ( Name = Exp ) = Exp", function()
      parser:parse("fn inc(x = 123) = x + 1")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        visib = nil,
        itype = nil,
        name = "inc",
        accessor = nil,
        rtype = nil,
        rvar = nil
      })
      assert(stmt.params):len(1)
      assert(stmt.params[1]):has({
        const = false,
        modifier = nil,
        name = "x",
        optional = true,
        type = nil,
      })
      assert(stmt.params[1].value:__tostring()):eq("123")
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("return (+ x 1)")
    end)

    test("fn Name ( Name = Exp ) = Xml", function()
      parser:parse("#!react\nfn inc(x = 123) = <X/>")

      parser:next() --#!react

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 2,
        col = 1,
        visib = nil,
        itype = nil,
        name = "inc",
        accessor = nil,
        rtype = nil,
        rvar = nil
      })
      assert(stmt.params):len(1)
      assert(stmt.params[1]):has({
        const = false,
        modifier = nil,
        name = "x",
        optional = true,
        type = nil,
      })
      assert(stmt.params[1].value:__tostring()):eq("123")
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("return (<X></X>)")
    end)

    test("fn Name ( Name : Name = Exp ) = Exp", function()
      parser:parse("fn inc(x : num = 123) = x + 1")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        visib = nil,
        itype = nil,
        name = "inc",
        accessor = nil,
        rtype = nil,
        rvar = nil
      })
      assert(stmt.params):len(1)
      assert(stmt.params[1]):has({
        const = false,
        modifier = nil,
        name = "x",
        optional = true,
        type = {
          kind = "be",
          type = {
            kind = "type",
            name = "num"
          }
        },
      })
      assert(stmt.params[1].value:__tostring()):eq("123")
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("return (+ x 1)")
    end)

    test("fn Name ( Name ? := Exp ) = Exp - error", function()
      parser:parse("fn inc(x ? := 123) = x + 1")
      assert(function() parser:next() end):raises("anonymous code (1, 12): ')' expected.")
    end)

    test("visibility fn Name ()", function(params)
      parser:parse(params[1] .. " fn myfn()")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        visib = params[1],
        async = false,
        itype = nil,
        name = "myfn",
        accessor = nil,
        rtype = nil,
        rvar = nil,
        params = {},
        body = {}
      })
    end):iter(
      {subtitle = " # export", params = "export"},
      {subtitle = " # pub", params = "pub"}
    )

    test("fn Name ()", function()
      parser:parse("fn myfn()")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        visib = nil,
        async = false,
        itype = nil,
        name = "myfn",
        accessor = nil,
        rtype = nil,
        rvar = nil,
        params = {},
        body = {}
      })
    end)

    test("@annot1\\nfn Name ()", function()
      parser:parse("@annot1\nfn myfn()")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 2,
        col = 1,
        visib = nil,
        async = false,
        itype = nil,
        name = "myfn",
        accessor = nil,
        rtype = nil,
        rvar = nil,
        params = {},
        body = {}
      })
      assert(stmt.annots:has("annot1")):eq(true)
    end)

    test("fn Name () -> self", function()
      parser:parse("fn myfn() -> self")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        visib = nil,
        async = false,
        itype = nil,
        name = "myfn",
        accessor = nil,
        rtype = nil,
        rvar = "self",
        params = {},
        body = {}
      })
    end)

    test("fn Name () -> Name", function()
      parser:parse("fn myfn() -> num")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        visib = nil,
        async = false,
        itype = nil,
        name = "myfn",
        accessor = nil,
        rtype = nil,
        rvar = "num",
        params = {},
        body = {}
      })
    end)

    test("fn Name () -> keyword", function()
      parser:parse("fn myfn() -> type")
      assert(function() parser:next() end):raises("anonymous code (1, 14): return value must be 'self' or a name.")
    end)

    test("fn Name () -> Name : Name", function()
      parser:parse("fn myfn() -> res:num")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        visib = nil,
        async = false,
        itype = nil,
        name = "myfn",
        accessor = nil,
        rtype = "num",
        rvar = "res",
        params = {},
        body = {}
      })
    end)

    test("fn Name . Name ()", function()
      parser:parse("fn MyType.myfn()")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        visib = "pub",
        async = false,
        itype = "MyType",
        name = "myfn",
        accessor = nil,
        rtype = nil,
        rvar = nil,
        params = {},
        body = {}
      })
    end)

    test("fn Name : Name ()", function()
      parser:parse("fn MyType:myfn()")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        visib = "intl",
        async = false,
        itype = "MyType",
        name = "myfn",
        accessor = nil,
        rtype = nil,
        rvar = nil,
        params = {},
        body = {}
      })
    end)

    test("fn Name ! Name ()", function()
      parser:parse("fn MyType!myfn()")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        visib = "pvt",
        async = false,
        itype = "MyType",
        name = "myfn",
        accessor = nil,
        rtype = nil,
        rvar = nil,
        params = {},
        body = {}
      })
    end)

    test("fn Name () = Exp", function()
      parser:parse("fn myfn() = 12 * 34")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        visib = nil,
        async = false,
        itype = nil,
        name = "myfn",
        accessor = nil,
        rtype = nil,
        rvar = nil,
        params = {}
      })
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("return (* 12 34)")
    end)

    test("fn Name () : Name = Exp", function()
      parser:parse("fn myfn() : num = 12 * 34")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        visib = nil,
        async = false,
        itype = nil,
        name = "myfn",
        rtype = "num",
        rvar = nil,
        accessor = nil,
        params = {}
      })
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("return (* 12 34)")
    end)

    test("fn Name ( Name , Name ) = Exp", function()
      parser:parse("fn sum(x, y) = x + y")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        visib = nil,
        async = false,
        itype = nil,
        name = "sum",
        accessor = nil,
        rtype = nil,
        rvar = nil,
        params = {
          {
            const = false,
            modifier = nil,
            name = "x",
            optional = false,
            type = nil,
            value = nil
          },
          {
            const = false,
            modifier = nil,
            name = "y",
            optional = false,
            type = nil,
            value = nil
          }
        }
      })
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("return (+ x y)")
    end)

    test("async fn Name() Body", function()
      parser:parse("async fn sum() return x + y")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        visib = nil,
        async = true,
        itype = nil,
        name = "sum",
        accessor = nil,
        rtype = nil,
        rvar = nil,
        params = {}
      })
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("return (+ x y)")
    end)

    test("async fn Name() -> var Body", function()
      parser:parse("async fn sum(x, y) -> r\n  r = x + y")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        visib = nil,
        async = true,
        itype = nil,
        name = "sum",
        accessor = nil,
        rtype = nil,
        rvar = nil,
        params = {
          {
            const = false,
            modifier = nil,
            name = "x",
            optional = false,
            type = nil,
            value = nil
          },
          {
            const = false,
            modifier = nil,
            name = "y",
            optional = false,
            type = nil,
            value = nil
          }
        }
      })
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(= r (+ x y))")
    end)
  end):tags("fn")

  ----------
  -- intf --
  ----------
  suite("intf", function()
    suite("errors", function()
      test(":type expected", function()
        parser:parse("export intf Param\n  x?")
        assert(function() parser:next() end):raises("anonymous code (2, 4): :type expected.")
      end)
    end)

    test("fields", function(params)
      parser:parse(params[1] .. ' intf Param\n  impl\n  b?:num\n  c:text\n  d:(text,"num")\n  e?:(text,num)')

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        type = "stmt",
        subtype = "intf",
        visib = params[1],
        name = "Param",
        fields = {
          {
            name = "impl",
            optional = false,
            type = nil
          },
          {
            name = "b",
            optional = true,
            type = {
              kind = "be",
              type = {
                kind = "type",
                name = "num"
              }
            }
          },
          {
            name = "c",
            optional = false,
            type = {
              kind = "be",
              type = {
                kind = "type",
                name = "text"
              }
            }
          },
          {
            name = "d",
            optional = false,
            type = {
              kind = "oneOf",
              types = {
                {
                  kind = "be",
                  type = {
                    kind = "type",
                    name = "text"
                  }
                },
                {
                  kind = "be",
                  type = {
                    kind = "value",
                    value = "num"
                  }
                }
              }
            }
          },
          {
            name = "e",
            optional = true,
            type = {
              kind = "oneOf",
              types = {
                {
                  kind = "be",
                  type = {
                    kind = "type",
                    name = "text"
                  }
                },
                {
                  kind = "be",
                  type = {
                    kind = "type",
                    name = "num"
                  }
                }
              }
            }
          }
        }
      })
      assert(stmt.annots):len(0)

      assert(parser:next()):isNil()
    end):iter(
      {subtitle="export", params="export"},
      {subtitle="pub", params="pub"}
    )

    test("methods", function()
      parser:parse("intf Interface\n  fn fn1()\n  fn fn2(x, y): z")
      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        type = "stmt",
        subtype = "intf",
        visib = nil,
        name = "Interface",
        fields = {},
        methods = {
          {name="fn1"},
          {name="fn2"}
        },
      })
      assert(stmt.annots):len(0)
      assert(parser:next()):isNil()
    end)

    test("intf", function()
      parser:parse("intf Param\n  x")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        type = "stmt",
        subtype = "intf",
        visib = nil,
        name = "Param",
        fields = {
          {name="x", optional=false, type=nil}
        }
      })
      assert(stmt.annots):len(0)

      assert(parser:next()):isNil()
    end)

    test("with base", function()
      parser:parse("intf Coord3D: Coord2D\n  z")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        type = "stmt",
        subtype = "intf",
        visib = nil,
        name = "Coord3D",
        base = "Coord2D",
        fields = {
          {name="z", optional=false, type=nil}
        }
      })
      assert(stmt.annots):len(0)

      assert(parser:next()):isNil()
    end)
  end):tags("intf")

  ------------
  -- struct --
  ------------
  suite("struct", function()
    suite("errors", function()
      test("var or const expected", function()
        parser:parse("struct MyStruct\n  pub x:num")
        assert(function() parser:next() end):raises("anonymous code (2, 7): 'const', 'var', 'async', 'fn', 'proc' or 'event' expected.")
      end)

      test("visibility expected", function()
        parser:parse("struct MyStruct\n  const x:num")
        assert(function() parser:next() end):raises("anonymous code (2, 3): 'pub', 'intl' or 'pvt' expected.")
      end)

      test("field or method expcted", function()
        parser:parse("struct MyStruct\n  x:num")
        assert(function() parser:next() end):raises("anonymous code (2, 3): field or method expected.")
      end)

      test("proc or fn expected when async", function()
        parser:parse("struct MyStruct\n  pub async x:num")
        assert(function() parser:next() end):raises("anonymous code (2, 13): 'proc' or 'fn' expected.")
      end)

      test("proc or fn expected when static", function()
        parser:parse("struct MyStruct\n  pub static x:num")
        assert(function() parser:next() end):raises("anonymous code (2, 14): 'proc' or 'fn' expected.")
      end)
    end)

    test("template", function()
      parser:parse("use Tmpl<T>\nexport struct MyStruct : Tmpl<MyT>\nx = 1", "./test/data/template/MyStruct.dog")
      assert(parser:next()):isTable():has({line=1, col=1, subtype=StmtType.USE})

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        visib = nil,
        name = "MyStruct",
        base = nil
      })
      assert(stmt.annots._items["Self"]):isNotNil()
      assert(parser:next()):isTable():has({line=3, col=1, type=SentType.EXP})
    end)

    test("@emitter struct Name", function()
      parser:parse("@emitter struct MyStruct")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 10,
        visib = nil,
        name = "MyStruct",
        base = "dogma.EventEmitter",
        fields = {}
      })
      assert(stmt.annots:has("emitter")):eq(true)
    end)

    test("export struct Name", function()
      parser:parse("export struct MyStruct")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        visib = "export",
        name = "MyStruct",
        base = nil,
        fields = {}
      })
    end)

    test("pub struct Name", function()
      parser:parse("pub struct MyStruct")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        visib = "pub",
        name = "MyStruct",
        base = nil,
        fields = {}
      })
    end)

    test("struct Name", function()
      parser:parse("struct Coord2D")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        visib = nil,
        name = "Coord2D",
        base = nil,
        fields = {},
      })
    end)

    test("struct Name : Name", function()
      parser:parse("struct Coord3D : Coord2D")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        visib = nil,
        name = "Coord3D",
        base = "Coord2D",
        fields = {},
      })
    end)

    test("struct Name : Name :: Name", function()
      parser:parse("struct Coord3D : Coord2D :: Mixin")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        visib = nil,
        name = "Coord3D",
        base = "Coord2D",
        mixins = {"Mixin"},
        margs = {"_"},
        fields = {},
      })
    end)

    test("struct Name : Name :: Name, Name", function()
      parser:parse("struct Coord3D : Coord2D :: Mixin1, Mixin2")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        visib = nil,
        name = "Coord3D",
        base = "Coord2D",
        mixins = {"Mixin1", "Mixin2"},
        margs = {"_"},
        fields = {},
      })
    end)

    suite("fields", function()
      suite("@strict", function()
        test("error with const - value expected", function()
          parser:parse("struct MyStruct\n  @strict pub const a")
          assert(function() parser:next() end):raises("anonymous code (2, 22): '=' expected.")
        end)

        test("error with var - value expected", function()
          parser:parse("struct MyStruct\n  @strict pub var a")
          assert(function() parser:next() end):raises("anonymous code (2, 20): '=' expected.")
        end)
      end)

      suite("list fields", function()
        test("type[...]", function()
          parser:parse("struct MyStruct\n  pub static var a: num[]\n  pub var b: bool[2]\n  pub var c: text[2..*]\n  pub var d: text[2..4]")
          stmt = parser:next()

          assert(stmt):isTable():has({
            line = 1,
            col = 1,
            visib = nil,
            name = "MyStruct",
            base = nil
          })
          assert(stmt.fields):len(4)

          assert(stmt.fields[1]):has({
            kind = "var",
            visib = "pub",
            static = true,
            name = "a",
            optional = false,
            type = {
              kind = "be",
              type = {
                kind = "type",
                name = "num",
                min = 0,
                max = nil
              }
            },
            value = nil
          })

          assert(stmt.fields[2]):has({
            kind = "var",
            visib = "pub",
            static = false,
            name = "b",
            optional = false,
            type = {
              kind = "be",
              type = {
                kind = "type",
                name = "bool",
                min = 2,
                max = 2
              }
            },
            value = nil
          })

          assert(stmt.fields[3]):has({
            kind = "var",
            visib = "pub",
            static = false,
            name = "c",
            optional = false,
            type = {
              kind = "be",
              type = {
                kind = "type",
                name = "text",
                min = 2,
                max = nil
              }
            },
            value = nil
          })

          assert(stmt.fields[4]):has({
            kind = "var",
            visib = "pub",
            static = false,
            name = "d",
            optional = false,
            type = {
              kind = "be",
              type = {
                kind = "type",
                name = "text",
                min = 2,
                max = 4
              }
            },
            value = nil
          })

          assert(parser:next()):isNil()
        end)

        test("(types)[...]", function()
          parser:parse("struct MyStruct\n  pub var a: (m,n)[]\n  pub var b: (m,n)[2]\n  pub var c: (m,n)[2..*]\n  pub var d: (m,n)[2..4]")
          stmt = parser:next()

          assert(stmt):isTable():has({
            line = 1,
            col = 1,
            visib = nil,
            name = "MyStruct",
            base = nil
          })
          assert(stmt.fields):len(4)

          assert(stmt.fields[1]):has({
            kind = "var",
            visib = "pub",
            name = "a",
            optional = false,
            type = {
              kind = "oneOf",
              types = {
                {kind = "be", type = {kind = "type", name = "m"}},
                {kind = "be", type = {kind = "type", name = "n"}}
              },
              min = 0,
              max = nil
            }
          })

          assert(stmt.fields[2]):has({
            kind = "var",
            visib = "pub",
            name = "b",
            optional = false,
            type = {
              kind = "oneOf",
              types = {
                {kind = "be", type = {kind = "type", name = "m"}},
                {kind = "be", type = {kind = "type", name = "n"}}
              },
              min = 2,
              max = 2
            }
          })

          assert(stmt.fields[3]):has({
            kind = "var",
            visib = "pub",
            name = "c",
            optional = false,
            type = {
              kind = "oneOf",
              types = {
                {kind = "be", type = {kind = "type", name = "m"}},
                {kind = "be", type = {kind = "type", name = "n"}}
              },
              min = 2,
              max = nil
            }
          })

          assert(stmt.fields[4]):has({
            kind = "var",
            visib = "pub",
            name = "d",
            optional = false,
            type = {
              kind = "oneOf",
              types = {
                {kind = "be", type = {kind = "type", name = "m"}},
                {kind = "be", type = {kind = "type", name = "n"}}
              },
              min = 2,
              max = 4
            }
          })

          assert(parser:next()):isNil()
        end)
      end)

      test("const fields", function()
        parser:parse('struct MyStruct\n  pub const a\n  pvt const b:num\n  intl const c:=0\n  pub const d:num=0\n  pub const e?:num')
        stmt = parser:next()
        assert(stmt):isTable():has({
          line = 1,
          col = 1,
          visib = nil,
          name = "MyStruct",
          base = nil
        })
        assert(stmt.fields):len(5)
        assert(stmt.fields[1]):has({
          kind = "const",
          visib = "pub",
          name = "a",
          optional = false,
          type = nil,
          value = nil
        })
        assert(stmt.fields[2]):has({
          kind = "const",
          visib = "pvt",
          name = "b",
          optional = false,
          type = {
            kind = "be",
            type = {
              kind = "type",
              name = "num"
            }
          },
          value=nil})
        assert(stmt.fields[3]):has({
          kind = "const",
          visib = "intl",
          name = "c",
          optional = true,
          type = {
            kind = "be",
            type = {
              kind = "type",
              name = "num"
            }
          }
        })
        assert(tostring(stmt.fields[3].value)):eq("0")
        assert(stmt.fields[4]):has({
          kind = "const",
          visib = "pub",
          name = "d",
          optional = true,
          type = {
            kind = "be",
            type = {
              kind = "type",
              name = "num"
            }
          }
        })
        assert(tostring(stmt.fields[4].value)):eq("0")
        assert(stmt.fields[5]):has({
          kind = "const",
          visib = "pub",
          name = "e",
          optional = true,
          type = {
            kind = "be",
            type = {
              kind = "type",
              name = "num"
            }
          },
          value = nil
        })
      end)

      test("@strict const fields", function()
        --(1) act
        parser:parse('struct MyStruct\n  @strict pub const a = 123\n  @strict pub const b: num = 123')
        stmt = parser:next()

        --(2) assert
        assert(stmt):isTable():has({
          line = 1,
          col = 1,
          visib = nil,
          name = "MyStruct",
          base = nil
        })
        assert(stmt.fields):len(2)

        local field = stmt.fields[1]
        assert(field):has({
          kind = "const",
          visib = "pub",
          name = "a",
          optional = true,
          type = nil,
        })
        assert(tostring(field.value)):eq("123")

        field = stmt.fields[2]
        assert(field):has({
          kind = "const",
          visib = "pub",
          name = "b",
          optional = true,
          type = {
            kind = "be",
            type = {kind = "type", name = "num"}
          }
        })
        assert(tostring(field.value)):eq("123")
      end)

      test("var fields", function()
        parser:parse('struct MyStruct\n  pub var a\n  pvt var b?\n  intl var c:num\n  pub var d?:num\n  pub var e:num=0\n  pub var f:=0')
        stmt = parser:next()
        assert(stmt):isTable():has({
          line = 1,
          col = 1,
          visib = nil,
          name = "MyStruct",
          base = nil
        })
        assert(stmt.fields):len(6)
        assert(stmt.fields[1]):has({
          kind = "var",
          visib = "pub",
          name = "a",
          optional = false,
          type = nil,
          value = nil
        })
        assert(stmt.fields[2]):has({
          kind = "var",
          visib = "pvt",
          name = "b",
          optional = true,
          type = nil,
          value = nil
        })
        assert(stmt.fields[3]):has({
          kind = "var",
          visib = "intl",
          name = "c",
          optional = false,
          type = {
            kind = "be",
            type = {
              kind = "type",
              name = "num"
            }
          },
          value = nil
        })
        assert(stmt.fields[4]):has({
          kind = "var",
          visib = "pub",
          name = "d",
          optional = true,
          type = {
            kind = "be",
            type = {
              kind = "type",
              name = "num"
            }
          },
          value = nil
        })
        assert(stmt.fields[5]):has({
          kind = "var",
          visib = "pub",
          name = "e",
          optional = true,
          type = {
            kind = "be",
            type = {
              kind = "type",
              name = "num"
            }
          }
        })
        assert(tostring(stmt.fields[5].value)):eq("0")
        assert(stmt.fields[6]):has({
          kind = "var",
          visib = "pub",
          name = "f",
          optional = true,
          type = {
            kind = "be",
            type = {
              kind = "type",
              name = "num"
            }
          }
        })
        assert(tostring(stmt.fields[6].value)):eq("0")
      end)

      test("annotations", function()
        parser:parse('struct MyStruct\n  @prop pub var a?\n  intl var b:num')
        stmt = parser:next()
        assert(stmt):isTable():has({
          line = 1,
          col = 1,
          visib = nil,
          name = "MyStruct",
          base = nil
        })
        assert(stmt.fields):len(2)
        assert(stmt.fields[1]):has({
          kind = "var",
          visib = "pub",
          name = "a",
          optional = true,
          type = nil,
          value = nil
        })
        assert(stmt.fields[1].annots:has("prop")):eq(true)
        assert(stmt.fields[2]):has({
          kind = "var",
          visib = "intl",
          name = "b",
          optional = false,
          type = {
            kind = "be",
            type = {
              kind = "type",
              name = "num"
            }
          },
          value = nil
        })
        assert(stmt.fields[2].annots:has("prop")):eq(false)
      end)
    end)

    suite("methods", function()
      test("fn", function()
        parser:parse('struct MyStruct\n  pub async fn method1()\n  @abstract\n  pvt fn method2()\n  intl fn method3()\n  pub static fn method4()\n  pvt static async fn method5()')
        stmt = parser:next()
        assert(stmt):isTable():has({
          line = 1,
          col = 1,
          visib = nil,
          name = "MyStruct",
          base = nil
        })
        assert(stmt.fields):len(0)
        assert(stmt.methods):len(5)

        --method1
        assert(stmt.methods[1]):has({
          name = "method1",
          itype = "MyStruct",
          visib = "pub",
          static = false,
          async = true,
          subtype = "fn",
          line = 2,
          col = 3,
        })

        --method2
        assert(stmt.methods[2]):has({
          name = "method2",
          itype = "MyStruct",
          visib = "pvt",
          static = false,
          async = false,
          subtype = "fn",
          line = 4,
          col = 3
        })
        assert(stmt.methods[2].annots._items["abstract"]):has({value="abstract"})

        --method3
        assert(stmt.methods[3]):has({
          name = "method3",
          itype = "MyStruct",
          visib = "intl",
          static = false,
          async = false,
          subtype = "fn",
          line = 5,
          col = 3
        })
        assert(stmt.methods[3].annots._items["abstract"]):isNil()

        --method4
        assert(stmt.methods[4]):has({
          name = "method4",
          itype = "MyStruct",
          visib = "pub",
          static = true,
          async = false,
          subtype = "fn",
          line = 6,
          col = 3
        })

        --method5
        assert(stmt.methods[5]):has({
          name = "method5",
          itype = "MyStruct",
          visib = "pvt",
          static = true,
          async = true,
          subtype = "fn",
          line = 7,
          col = 3
        })
      end)

      test("proc", function()
        parser:parse('struct MyStruct\n  pub async proc method1()\n  @abstract\n  pvt proc method2()\n  intl proc method3()\n  pub static proc method4()\n  pvt static async proc method5()')
        stmt = parser:next()
        assert(stmt):isTable():has({
          line = 1,
          col = 1,
          visib = nil,
          name = "MyStruct",
          base = nil
        })
        assert(stmt.fields):len(0)
        assert(stmt.methods):len(5)

        --method1
        assert(stmt.methods[1]):has({
          name = "method1",
          itype = "MyStruct",
          visib = "pub",
          static = false,
          async = true,
          subtype = "proc",
          line = 2,
          col = 3
        })

        --method2
        assert(stmt.methods[2]):has({
          name = "method2",
          itype = "MyStruct",
          visib = "pvt",
          static = false,
          async = false,
          subtype = "proc",
          line = 4,
          col = 3
        })
        assert(stmt.methods[2].annots._items["abstract"]):has({value="abstract"})

        --method3
        assert(stmt.methods[3]):has({
          name = "method3",
          itype = "MyStruct",
          visib = "intl",
          static = false,
          async = false,
          subtype = "proc",
          line = 5,
          col = 3
        })
        assert(stmt.methods[3].annots._items["abstract"]):isNil()

        --method4
        assert(stmt.methods[4]):has({
          name = "method4",
          itype = "MyStruct",
          visib = "pub",
          static = true,
          async = false,
          subtype = "proc",
          line = 6,
          col = 3
        })
        assert(stmt.methods[4].annots._items["abstract"]):isNil()

        --method5
        assert(stmt.methods[5]):has({
          name = "method5",
          itype = "MyStruct",
          visib = "pvt",
          static = true,
          async = true,
          subtype = "proc",
          line = 7,
          col = 3
        })
        assert(stmt.methods[5].annots._items["abstract"]):isNil()
      end)
    end)

    suite("events", function()
      test("without type", function()
        parser:parse("@emitter\nstruct MyStruct\n  pub event start")
        stmt = parser:next()
        assert(stmt):isTable():has({
          line = 2,
          col = 1,
          visib = nil,
          name = "MyStruct",
          base = nil
        })
        assert(stmt.events):isArray():eq({
          {
            visib = "pub",
            name = "start",
            type = nil
          }
        })
      end)

      test("with type", function()
        parser:parse("@emitter\nstruct MyStruct\n  pub event start: {task: Task, ts: timestamp}")
        stmt = parser:next()
        assert(stmt):isTable():has({
          line = 2,
          col = 1,
          visib = nil,
          name = "MyStruct",
          base = nil
        })
        assert(stmt.events):isArray():eq({
          {
            visib = "pub",
            name = "start",
            type = {
              kind = "map",
              fields = {
                {
                  name = "task",
                  type = {
                    kind = "be",
                    type = {name = "Task", kind = "type"}
                  },
                  mandatory = true
                },

                {
                  name = "ts",
                  type = {
                    kind = "be",
                    type = {name = "timestamp", kind = "type"}
                  },
                  mandatory = true
                }
              }
            }
          }
        })
      end)
    end)
  end)

  -------------
  -- typedef --
  -------------
  suite("typedef", function()
    test("type Name = Name", function()
      parser:parse("type A = B")
      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        visib = nil,
        name = "A",
        types = {"B"},
        min = nil,
        max = nil
      })
      assert(parser:next()):isNil()
    end)

    test("type Name = Name[]", function()
      parser:parse("type A = B[]")
      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        visib = nil,
        name = "A",
        types = {"B"},
        min = 0,
        max = nil
      })
      assert(parser:next()):isNil()
    end)

    test("type Name = Name[Num]", function()
      parser:parse("type A = B[2]")
      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        visib = nil,
        name = "A",
        types = {"B"},
        min = 2,
        max = 2
      })
      assert(parser:next()):isNil()
    end)

    test("type Name = Name[Num..*]", function()
      parser:parse("type A = B[2..*]")
      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        visib = nil,
        name = "A",
        types = {"B"},
        min = 2,
        max = nil
      })
      assert(parser:next()):isNil()
    end)

    test("type Name = Name[Num..Num]", function()
      parser:parse("type A = B[2..4]")
      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        visib = nil,
        name = "A",
        types = {"B"},
        min = 2,
        max = 4
      })
      assert(parser:next()):isNil()
    end)

    test("type Name = (Name,Name)", function()
      parser:parse("type A = (B, C)")
      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        visib = nil,
        name = "A",
        types = {"B", "C"},
        min = nil,
        max = nil
      })
      assert(parser:next()):isNil()
    end)

    test("type Name = (Name,Name)[]", function()
      parser:parse("type A = (B, C)[]")
      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        visib = nil,
        name = "A",
        types = {"B", "C"},
        min = 0,
        max = nil
      })
      assert(parser:next()):isNil()
    end)
  end)

  ----------
  -- type --
  ----------
  suite("type", function()
    test("@emitter\\ntype Name()", function()
      parser:parse("@emitter\ntype Abc()")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 2,
        col = 1,
        visib = nil,
        name = "Abc",
        params = {},
        base = "dogma.EventEmitter",
        bargs = nil,
        body = {},
        catch = nil,
        finally = nil
      })
      assert(stmt.annots:has("emitter")):eq(true)
    end)

    test("@annot1\\ntype Name()", function()
      parser:parse("@annot1\ntype Coord2D()")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 2,
        col = 1,
        visib = nil,
        name = "Coord2D",
        params = {},
        base = nil,
        bargs = nil,
        body = {},
        catch = nil,
        finally = nil
      })
      assert(stmt.annots:has("annot1")):eq(true)
    end)

    test("type Name", function()
      parser:parse("type Coord2D")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        visib = nil,
        name = "Coord2D",
        params = {},
        base = nil,
        bargs = nil,
        body = nil,
        catch = nil,
        finally = nil
      })
    end)

    test("type Name with fields", function()
      parser:parse('type Coord2D\n  pub const x:=0\n  var y?:num=0\n  pvt const ctype:="2D"')
      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        visib = nil,
        name = "Coord2D",
        base = nil,
        bargs = nil,
        body = nil,
        catch = nil,
        finally = nil
      })
      assert(stmt.params):len(3)
      assert(stmt.params[1]):isTable():has({
        const = true,
        modifier = ".",
        name = "x",
        type = {
          kind = "be",
          type = {
            kind = "type",
            name = "num"
          }
        },
        optional = false,
      })
      assert(tostring(stmt.params[1].value)):eq("0")
      assert(stmt.params[2]):isTable():has({
        const = false,
        modifier = ".",
        type = {
          kind = "be",
          type = {
            kind = "type",
            name = "num"
          }
        },
        optional = true,
        value = nil
      })
      assert(tostring(stmt.params[2].value)):eq("0")
      assert(stmt.params[3]):isTable():has({
        const = true,
        modifier = ":",
        name = "ctype",
        type = {
          kind = "be",
          type = {
            kind = "type",
            name = "text"
          }
        },
        optional = false,
      })
      assert(tostring(stmt.params[3].value)):eq("2D")
    end)

    test("type Name - error: const or var expected", function()
      parser:parse('type Coord2D\n  pub x:=0\n  var y?:num=0\n  pvt const ctype:="2D"')
      assert(function() parser:next() end):raises("anonymous code (2, 7): var or const expected.")
    end)

    test("type Name()", function()
      parser:parse("type Coord2D()")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        visib = nil,
        name = "Coord2D",
        params = {},
        base = nil,
        bargs = nil,
        body = {},
        catch = nil,
        finally = nil
      })
    end)

    test("visibility type Name()", function(params)
      parser:parse(params[1] .. " type Coord2D()")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        visib = params[1],
        name = "Coord2D",
        params = {},
        base = nil,
        bargs = nil,
        body = {},
        catch = nil,
        finally = nil
      })
    end):iter(
      {subtitle = " # export", params = "export"},
      {subtitle = " # pub", params = "pub"}
    )

    test("type Name(params)", function()
      parser:parse("type Coord2D(x, y)\n  .x = x\n  .y = y")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        visib = nil,
        name = "Coord2D",
        base = nil,
        bargs = nil,
        catch = nil,
        finally = nil
      })
      assert(stmt.params):len(2)
      assert(stmt.params[1]):has({name = "x"})
      assert(stmt.params[2]):has({name = "y"})
      assert(stmt.body):len(2)
      assert(stmt.body[1]:__tostring()):eq("(= (. x) x)")
      assert(stmt.body[2]:__tostring()):eq("(= (. y) y)")
    end)

    test("type Name() : Name", function()
      parser:parse("type Coord3D() : Coord2D")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        visib = nil,
        name = "Coord3D",
        params = {},
        base = "Coord2D",
        bargs = nil,
        body = {},
        catch = nil,
        finally = nil
      })
    end)

    test("type Name() : Name.Name", function()
      parser:parse("type Coord3D() : Coord.D2")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        visib = nil,
        name = "Coord3D",
        params = {},
        base = "Coord.D2",
        bargs = nil,
        body = {},
        catch = nil,
        finally = nil
      })
    end)

    test("type Name(params) : Name(args)", function()
      parser:parse("type Coord3D(x, y, .z) : Coord2D(x, y)")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        visib = nil,
        name = "Coord3D",
        base = "Coord2D",
        body = {},
        catch = nil,
        finally = nil
      })
      assert(stmt.params):len(3)
      assert(stmt.params[1]):has({name = "x", modifier = nil})
      assert(stmt.params[2]):has({name = "y", modifier = nil})
      assert(stmt.params[3]):has({name = "z", modifier = "."})
      assert(stmt.bargs):len(2)
      assert(stmt.bargs[1]:__tostring()):eq("x")
      assert(stmt.bargs[2]:__tostring()):eq("y")
    end)

    test("type Name(params) : Name(...)", function()
      parser:parse("type Coord3D(x, y, .z) : Coord2D(...)")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        visib = nil,
        name = "Coord3D",
        base = "Coord2D",
        body = {},
        catch = nil,
        finally = nil
      })
      assert(stmt.params):len(3)
      assert(stmt.params[1]):has({name = "x", modifier = nil})
      assert(stmt.params[2]):has({name = "y", modifier = nil})
      assert(stmt.params[3]):has({name = "z", modifier = "."})
      assert(stmt.bargs):len(1)
      assert(stmt.bargs[1]):eq("...")
    end)

    test("type Name(...) : Name(...)", function()
      parser:parse("type Coord3D(...) : Coord2D(...)")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        visib = nil,
        name = "Coord3D",
        base = "Coord2D",
        mixin = nil,
        margs = nil,
        body = {},
        catch = nil,
        finally = nil
      })
      assert(stmt.params):len(0)
      assert(stmt.bargs):len(1)
      assert(stmt.bargs[1]):eq("...")
    end)

    suite("mixin", function()
      test("@mixable type Name", function()
        parser:parse("@mixable type Startable")

        stmt = parser:next()
        assert(stmt):isTable():has({
          line = 1,
          col = 10,
          visib = nil,
          name = "Startable",
          base = nil,
          body = nil,
          catch = nil,
          finally = nil
        })
        assert(stmt.annots:has("mixable")):eq(true)
        assert(stmt.params):len(0)
        assert(stmt.bargs):isNil()
      end)

      test("type Name :: Mixin()", function()
        parser:parse("type Server() :: Startable()")

        stmt = parser:next()
        assert(stmt):isTable():has({
          line = 1,
          col = 1,
          visib = nil,
          name = "Server",
          base = nil,
          bargs = nil,
          mixin = "Startable",
          margs = {},
          body = nil,
          catch = nil,
          finally = nil
        })
        assert(stmt.params):len(0)
        assert(stmt.bargs):isNil()
      end)

      test("type Name() :: Mixin(...)", function()
        parser:parse("type Server() :: Startable(...)")

        stmt = parser:next()
        assert(stmt):isTable():has({
          line = 1,
          col = 1,
          visib = nil,
          name = "Server",
          base = nil,
          bargs = nil,
          mixin = "Startable",
          margs = {"..."},
          body = nil,
          catch = nil,
          finally = nil
        })
        assert(stmt.params):len(0)
        assert(stmt.bargs):isNil()
      end)

      test("type Name :: Mixin(args)", function()
        parser:parse("type Server() :: Startable(one, two)")

        stmt = parser:next()
        assert(stmt):isTable():has({
          line = 1,
          col = 1,
          visib = nil,
          name = "Server",
          base = nil,
          bargs = nil,
          mixin = "Startable",
          body = nil,
          catch = nil,
          finally = nil
        })
        assert(stmt.params):len(0)
        assert(stmt.bargs):isNil()
        assert(stmt.margs):len(2)
        assert(tostring(stmt.margs[1])):eq("one")
        assert(tostring(stmt.margs[2])):eq("two")
      end)
    end):tags("mixin")
  end):tags("type")

  -----------
  -- async --
  -----------
  suite("async", function()
    test("async Exp", function()
      parser:parse("async 1+2")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        subtype = StmtType.ASYNC,
        opts = {},
        catch = nil
      })
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring(("+ 1 2")))
    end)

    test("async with {invalid=Exp} Exp", function()
      parser:parse("async with {invalid=250+250} 1+2")
      assert(function() parser:next() end):raises("delay or repeat expected.")
    end)

    test("async with {delay=Exp} Exp", function()
      parser:parse("async with {delay=250+250} 1+2")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        subtype = StmtType.ASYNC,
        catch = nil
      })
      assert(tostring(stmt.opts.delay)):eq("(+ 250 250)")
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring(("+ 1 2")))
    end)

    test("async with {delay=Exp, repeat=Exp} Exp", function()
      parser:parse("async with {delay=250+250, repeat=10+20} 1+2")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        subtype = StmtType.ASYNC,
        catch = nil
      })
      assert(tostring(stmt.opts.delay)):eq("(+ 250 250)")
      assert(tostring(stmt.opts["repeat"])):eq("(+ 10 20)")
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring(("+ 1 2")))
    end)

    test("async with {delay} Exp", function()
      parser:parse("async with {delay} 1+2")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        subtype = StmtType.ASYNC,
        catch = nil
      })
      assert(tostring(stmt.opts.delay)):eq("delay")
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring(("+ 1 2")))
    end)

    test("async\\n Exp\\n Exp", function()
      parser:parse("async\n 1+2\n 3+4")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        subtype = StmtType.ASYNC,
        opts = {},
        catch = nil
      })
      assert(stmt.body):len(2)
      assert(stmt.body[1]:__tostring(("+ 1 2")))
      assert(stmt.body[2]:__tostring(("+ 3 4")))
    end)

    test("async\\n Exp\\ncatch\\n Exp", function()
      parser:parse("async\n 1+2\ncatch\n 3+4")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        subtype = StmtType.ASYNC,
        opts = {}
      })
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring(("+ 1 2")))
      assert(stmt.catch.body):len(1)
      assert(stmt.catch.body[1]:__tostring(("+ 3 4")))
    end)
  end):tags("async")

  --------
  -- if --
  --------
  suite("if", function()
    test("if Exp then Exp", function()
      parser:parse("if x == 123 then x = 321")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        subtype = StmtType.IF,
        decl = nil,
        elif = nil,
        el = nil
      })
      assert(stmt.cond:__tostring()):eq("(== x 123)")
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(= x 321)")
    end)

    test("if name := Exp then Exp", function()
      parser:parse("if x := 123 then x = 321")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        subtype = StmtType.IF,
        decl = "var",
        elif = nil,
        el = nil
      })
      assert(tostring(stmt.value)):eq("(= x 123)")
      assert(tostring(stmt.cond)):eq("x")
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(= x 321)")
    end)

    test("if name ::= Exp then Exp", function()
      parser:parse("if x ::= 123 then x = 321")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        subtype = StmtType.IF,
        decl = "const",
        elif = nil,
        el = nil
      })
      assert(tostring(stmt.value)):eq("(= x 123)")
      assert(tostring(stmt.cond)):eq("x")
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(= x 321)")
    end)

    test("if var Exp then Exp", function()
      parser:parse("if var x = 123 then x = 321")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        subtype = StmtType.IF,
        decl = "var",
        elif = nil,
        el = nil
      })
      assert(tostring(stmt.value)):eq("(= x 123)")
      assert(tostring(stmt.cond)):eq("x")
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(= x 321)")
    end)

    test("if var Exp; then Exp", function()
      parser:parse("if var x = 123; x then x = 321")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        subtype = StmtType.IF,
        decl = "var",
        elif = nil,
        el = nil
      })
      assert(tostring(stmt.value)):eq("(= x 123)")
      assert(tostring(stmt.cond)):eq("x")
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(= x 321)")
    end)

    test("if const Exp; then Exp", function()
      parser:parse("if const x = 123; x then x = 321")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        subtype = StmtType.IF,
        decl = "const",
        elif = nil,
        el = nil
      })
      assert(tostring(stmt.value)):eq("(= x 123)")
      assert(tostring(stmt.cond)):eq("x")
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(= x 321)")
    end)

    test("if Exp; Exp then Exp", function()
      parser:parse("if x == 123; x then x = 321")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        subtype = StmtType.IF,
        decl = nil,
        elif = nil,
        el = nil
      })
      assert(tostring(stmt.value)):eq("(== x 123)")
      assert(tostring(stmt.cond)):eq("x")
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(= x 321)")
    end)

    test("if Exp then Stmt", function()
      parser:parse("if x == 123 then return 1+2")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        subtype = StmtType.IF,
        decl = nil,
        elif = nil,
        el = nil
      })
      assert(stmt.cond:__tostring()):eq("(== x 123)")
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("return (+ 1 2)")
    end)

    test("if Exp then Exp else Exp", function()
      parser:parse("if x == 123 then x = 321 else x = 135")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        subtype = StmtType.IF,
        decl = nil,
        elif = nil,
      })
      assert(stmt.cond:__tostring()):eq("(== x 123)")
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(= x 321)")
      assert(stmt.el):len(1)
      assert(stmt.el[1]:__tostring()):eq("(= x 135)")
    end)

    test("if Exp then Exp\\nelse Exp", function()
      parser:parse("if x == 123 then x = 321\nelse x = 135")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        subtype = StmtType.IF,
        decl = nil,
        elif = nil,
      })
      assert(stmt.cond:__tostring()):eq("(== x 123)")
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(= x 321)")
      assert(stmt.el):len(1)
      assert(stmt.el[1]:__tostring()):eq("(= x 135)")
    end)

    test("if Exp then\\n Exp", function()
      parser:parse("if x == 123 then\n x = 321")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        subtype = StmtType.IF,
        decl = nil,
        elif = nil,
        el = nil
      })
      assert(stmt.cond:__tostring()):eq("(== x 123)")
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(= x 321)")
    end)

    test("if Exp then\\n Exp\\nelse\\n Exp", function()
      parser:parse("if x == 123 then\n x = 321\nelse\n x = 135")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        subtype = StmtType.IF,
        decl = nil,
        elif = nil,
      })
      assert(stmt.cond:__tostring()):eq("(== x 123)")
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(= x 321)")
      assert(stmt.el):len(1)
      assert(stmt.el[1]:__tostring()):eq("(= x 135)")
    end)

    test("if Exp then\\n Exp\\nelse if Exp then\\n Exp", function()
      parser:parse("if x == 111 then\n x = 222\nelse if x == 333 then\n x = 444")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        subtype = StmtType.IF,
        decl = nil,
        el = nil
      })
      assert(stmt.cond:__tostring()):eq("(== x 111)")
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(= x 222)")
      assert(stmt.elif):len(1)
      assert(stmt.elif[1].cond:__tostring()):eq("(== x 333)")
      assert(stmt.elif[1].body):len(1)
      assert(stmt.elif[1].body[1]:__tostring()):eq("(= x 444)")
    end)

    test("if Exp then Exp\\nelse if Exp then Exp", function()
      parser:parse("if x == 111 then x = 222\nelse if x == 333 then x = 444")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        subtype = StmtType.IF,
        decl = nil,
        el = nil
      })
      assert(stmt.cond:__tostring()):eq("(== x 111)")
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(= x 222)")
      assert(stmt.elif):len(1)
      assert(stmt.elif[1].cond:__tostring()):eq("(== x 333)")
      assert(stmt.elif[1].body):len(1)
      assert(stmt.elif[1].body[1]:__tostring()):eq("(= x 444)")
    end)

    test("if Exp then\\n Exp\\nelse if Exp then\\n Exp\\nelse if Exp then\\n Exp", function()
      parser:parse("if x == 111 then\n x = 222\nelse if x == 333 then\n x = 444\nelse if x == 555 then\n x = 666")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        subtype = StmtType.IF,
        decl = nil,
        el = nil
      })
      assert(stmt.cond:__tostring()):eq("(== x 111)")
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(= x 222)")
      assert(stmt.elif):len(2)
      assert(stmt.elif[1].cond:__tostring()):eq("(== x 333)")
      assert(stmt.elif[1].body):len(1)
      assert(stmt.elif[1].body[1]:__tostring()):eq("(= x 444)")
      assert(stmt.elif[2].cond:__tostring()):eq("(== x 555)")
      assert(stmt.elif[2].body):len(1)
      assert(stmt.elif[2].body[1]:__tostring()):eq("(= x 666)")
    end)

    test("if Exp then Exp\\nelse if Exp then Exp\\nelse if Exp then Exp", function()
      parser:parse("if x == 111 then x = 222\nelse if x == 333 then x = 444\nelse if x == 555 then x = 666")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        subtype = StmtType.IF,
        decl = nil,
        el = nil
      })
      assert(stmt.cond:__tostring()):eq("(== x 111)")
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(= x 222)")
      assert(stmt.elif):len(2)
      assert(stmt.elif[1].cond:__tostring()):eq("(== x 333)")
      assert(stmt.elif[1].body):len(1)
      assert(stmt.elif[1].body[1]:__tostring()):eq("(= x 444)")
      assert(stmt.elif[2].cond:__tostring()):eq("(== x 555)")
      assert(stmt.elif[2].body):len(1)
      assert(stmt.elif[2].body[1]:__tostring()):eq("(= x 666)")
    end)

    test("if Exp then\\n Exp\\nelse if Exp then\\n Exp\\nelse\\n Exp", function()
      parser:parse("if x == 111 then\n x = 222\nelse if x == 333 then\n x = 444\nelse x = 555")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        subtype = StmtType.IF,
        decl = nil

      })
      assert(stmt.cond:__tostring()):eq("(== x 111)")
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(= x 222)")
      assert(stmt.elif):len(1)
      assert(stmt.elif[1].cond:__tostring()):eq("(== x 333)")
      assert(stmt.elif[1].body):len(1)
      assert(stmt.elif[1].body[1]:__tostring()):eq("(= x 444)")
      assert(stmt.el):len(1)
      assert(stmt.el[1]:__tostring()):eq("(= x 555)")
    end)

    test("if Exp then Exp\\nelse if Exp then Exp\\nelse Exp", function()
      parser:parse("if x == 111 then x = 222\nelse if x == 333 then x = 444\nelse x = 555")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        subtype = StmtType.IF,
        decl = nil

      })
      assert(stmt.cond:__tostring()):eq("(== x 111)")
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(= x 222)")
      assert(stmt.elif):len(1)
      assert(stmt.elif[1].cond:__tostring()):eq("(== x 333)")
      assert(stmt.elif[1].body):len(1)
      assert(stmt.elif[1].body[1]:__tostring()):eq("(= x 444)")
      assert(stmt.el):len(1)
      assert(stmt.el[1]:__tostring()):eq("(= x 555)")
    end)

    test("  if Exp then\\n Exp\\nelse", function()
      parser:parse("  if x == 111 then\n    x = 222\nelse\n  x = 333")
      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 3,
        subtype = StmtType.IF,
        decl = nil

      })
      assert(stmt.cond:__tostring()):eq("(== x 111)")
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(= x 222)")
      assert(stmt.elif):isNil()
      assert(stmt.el):isNil()
    end)

    test("  if Exp then\\n Exp\\n  else if Exp then\\n    Exp\\nelse", function()
      parser:parse("  if x == 111 then\n    x = 222\n  else if x == 333 then\n    x = 444\nelse\n  x = 555")
      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 3,
        subtype = StmtType.IF,
        decl = nil

      })
      assert(stmt.cond:__tostring()):eq("(== x 111)")
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(= x 222)")
      assert(stmt.elif):len(1)
      assert(stmt.elif[1].cond:__tostring()):eq("(== x 333)")
      assert(stmt.elif[1].body):len(1)
      assert(stmt.elif[1].body[1]:__tostring()):eq("(= x 444)")
      assert(stmt.el):isNil()
    end)
  end):tags("if")

  ---------
  -- pub --
  ---------
  suite("pub", function()
    suite("error", function()
      test("pub - error, expecting text or name", function()
        parser:parse("pub")
        assert(function() parser:next() end):raises("anonymous code (1, 4): literal text or name expected.")
      end)

      test("pub Name, - error, expecting text or name", function()
        parser:parse("pub x,")
        assert(function() parser:next() end):raises("anonymous code (1, 7): literal text or name expected.")
      end)

      test("pub Name Name - error", function()
        parser:parse("pub Item1 Item2")
        assert(function() parser:next() end):raises("anonymous code (1, 11): comma or newline expected.")
      end)

      test("pub 123 - error", function()
        parser:parse("pub 123")
        assert(function() parser:next() end):raises("anonymous code (1, 5): literal text or name expected.")
      end)
    end)

    test("pub Name", function()
      parser:parse("pub Item")
      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        subtype = StmtType.PUB,
        items = {{type = "pub", value = "Item"}}
      })
    end)

    test("pub Text", function()
      parser:parse('pub "Item"')
      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        subtype = StmtType.PUB,
        items = {{type = "use", value = {path = "Item", name = "Item"}}}
      })
    end)

    test("pub Name, Text", function()
      parser:parse('pub Item1, "Item2"')
      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        subtype = StmtType.PUB,
        items = {{type = "pub", value = "Item1"}, {type = "use", value = {path = "Item2", name = "Item2"}}}
      })
    end)

    test("pub (\\nName1, Name2\\nText\\n)", function()
      parser:parse('pub (\nItem1, Item2\n"Item3"\n)')
      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        subtype = StmtType.PUB,
        items = {
          {type="pub", value="Item1"},
          {type="pub", value="Item2"},
          {type="use", value={path="Item3", name="Item3"}}
        }
      })
    end)

    test("pub use(\\nName1 as Name11, Name2\\nText\\n)", function()
      parser:parse('pub use(\nItem11 as Item1, Item2\nItem3 = "Item33"\ndir/Item4\n)')
      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        subtype = StmtType.PUB,
        items = {
          {type = "use", value = {path = "./Item11", name = "Item1"}},
          {type = "use", value = {path = "./Item2", name = "Item2"}},
          {type = "use", value = {path = "Item33", name = "Item3"}},
          {type = "use", value = {path = "./dir/Item4", name = "Item4"}}
        }
      })
    end)
  end):tags("pub")

  ------------
  -- export --
  ------------
  suite("export", function()
    test("export", function()
      parser:parse("export")
      assert(function() parser:next() end):raises("anonymous code (1, 7): incomplete expression.")
    end)

    test("export Exp", function()
      parser:parse("export 1+2")
      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        subtype = StmtType.EXPORT
      })
      assert(tostring(stmt.exp)):eq("(+ 1 2)")
    end)
  end):tags("export")

  ----------
  -- with --
  ----------
  suite("with", function()
    test("with", function()
      parser:parse("with")
      assert(function() parser:next() end):raises("anonymous code (1, 5): incomplete expression.")
    end)

    test("with var := Exp do", function()
      parser:parse("with x := 1+2 do")
      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        ofType = false,
        var = {name = "x", kind = "var"},
        subtype = StmtType.WITH,
        ifs = {},
        els = nil
      })
      assert(parser:next()):isNil()
    end)

    test("with var ::= Exp do", function()
      parser:parse("with x ::= 1+2 do")
      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        ofType = false,
        var = {name = "x", kind = "const"},
        subtype = StmtType.WITH,
        ifs = {},
        els = nil
      })
      assert(parser:next()):isNil()
    end)

    test("with Exp do", function()
      parser:parse("with 1+2 do")
      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        ofType = false,
        var = nil,
        subtype = StmtType.WITH,
        ifs = {},
        els = nil
      })
      assert(parser:next()):isNil()
    end)

    test("with Exp do if Exp then - with one sentence", function()
      parser:parse("with 1+2 do\n  if 1 then x")
      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        ofType = false,
        subtype = StmtType.WITH,
        var = nil,
        els = nil
      })
      assert(tostring(stmt.value)):eq("(+ 1 2)")
      assert(stmt.ifs):len(1)
      assert(tostring(stmt.ifs[1].conds[1])):eq("1")
      assert(stmt.ifs[1].body):len(1)
      assert(tostring(stmt.ifs[1].body[1])):eq("x")
      assert(parser:next()):isNil()
    end)

    test("with Exp do if Exp, Exp then", function()
      parser:parse("with 1+2 do\n  if 1, 2 then x")
      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        ofType = false,
        subtype = StmtType.WITH,
        var = nil,
        els = nil
      })
      assert(tostring(stmt.value)):eq("(+ 1 2)")
      assert(stmt.ifs):len(1)
      assert(tostring(stmt.ifs[1].conds[1])):eq("1")
      assert(tostring(stmt.ifs[1].conds[2])):eq("2")
      assert(stmt.ifs[1].body):len(1)
      assert(tostring(stmt.ifs[1].body[1])):eq("x")
      assert(parser:next()):isNil()
    end)

    test("with Exp do if Exp then - with several sentences", function()
      parser:parse("with 1+2 do\n  if 1 then x\n    y")
      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        ofType = false,
        subtype = StmtType.WITH,
        els = nil
      })
      assert(tostring(stmt.value)):eq("(+ 1 2)")
      assert(stmt.ifs):len(1)
      assert(tostring(stmt.ifs[1].conds[1])):eq("1")
      assert(stmt.ifs[1].body):len(2)
      assert(tostring(stmt.ifs[1].body[1])):eq("x")
      assert(tostring(stmt.ifs[1].body[2])):eq("y")
      assert(parser:next()):isNil()
    end)

    test("with Exp do - with several if", function()
      parser:parse("with 1+2 do\n  if 1 then x\n  if 2 then y")
      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        ofType = false,
        subtype = StmtType.WITH,
        els = nil
      })
      assert(tostring(stmt.value)):eq("(+ 1 2)")
      assert(stmt.ifs):len(2)
      assert(tostring(stmt.ifs[1].conds[1])):eq("1")
      assert(stmt.ifs[1].body):len(1)
      assert(tostring(stmt.ifs[1].body[1])):eq("x")
      assert(tostring(stmt.ifs[2].conds[1])):eq("2")
      assert(stmt.ifs[2].body):len(1)
      assert(tostring(stmt.ifs[2].body[1])):eq("y")
      assert(parser:next()):isNil()
    end)

    test("with Exp do - with if and else", function()
      parser:parse("with 1+2 do\n  if 1 then x\n  if 2 then y\n  else z\n")
      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        ofType = false,
        subtype = StmtType.WITH
      })
      assert(tostring(stmt.value)):eq("(+ 1 2)")
      assert(stmt.ifs):len(2)
      assert(tostring(stmt.ifs[1].conds[1])):eq("1")
      assert(stmt.ifs[1].body):len(1)
      assert(tostring(stmt.ifs[1].body[1])):eq("x")
      assert(tostring(stmt.ifs[2].conds[1])):eq("2")
      assert(stmt.ifs[2].body):len(1)
      assert(tostring(stmt.ifs[2].body[1])):eq("y")
      assert(stmt.els):isNotNil()
      assert(stmt.els):len(1)
      assert(tostring(stmt.els[1])):eq("z")
      assert(parser:next()):isNil()
    end)

    test("full with", function()
      parser:parse("with 1+2 do\n  if 1 then x\n    y\n  if 2 then a\n    b\n  else 1\n    2\ncatch e\n  print(y)\nfinally\n  print(z)\nnop")
      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        ofType = false,
        subtype = StmtType.WITH
      })
      assert(tostring(stmt.value)):eq("(+ 1 2)")
      assert(stmt.ifs):len(2)
      assert(tostring(stmt.ifs[1].conds[1])):eq("1")
      assert(stmt.ifs[1].body):len(2)
      assert(tostring(stmt.ifs[1].body[1])):eq("x")
      assert(tostring(stmt.ifs[1].body[2])):eq("y")
      assert(tostring(stmt.ifs[2].conds[1])):eq("2")
      assert(stmt.ifs[2].body):len(2)
      assert(tostring(stmt.ifs[2].body[1])):eq("a")
      assert(tostring(stmt.ifs[2].body[2])):eq("b")
      assert(stmt.els):isNotNil()
      assert(stmt.els):len(2)
      assert(tostring(stmt.els[1])):eq("1")
      assert(tostring(stmt.els[2])):eq("2")
      assert(stmt.catch):isTable()
      assert(stmt.catch.var):eq("e")
      assert(stmt.catch.body):len(1)
      assert(stmt.catch.body[1]:__tostring()):eq("(call print y)")
      assert(stmt.finally):isTable()
      assert(stmt.finally.body):len(1)
      assert(stmt.finally.body[1]:__tostring()):eq("(call print z)")
      assert(parser:next()):isTable():has({
        line = 12,
        col = 1,
        type = StmtType.EXP
      })
    end)

    test("full with type()", function()
      parser:parse("with type(x) do\n  if text then x\n    y\n  if num then a\n    b\n  else 1\n    2\ncatch e\n  print(y)\nfinally\n  print(z)\nnop")
      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        ofType = true,
        subtype = StmtType.WITH
      })
      assert(tostring(stmt.value)):eq("x")
      assert(stmt.ifs):len(2)
      assert(tostring(stmt.ifs[1].conds[1])):eq("text")
      assert(stmt.ifs[1].body):len(2)
      assert(tostring(stmt.ifs[1].body[1])):eq("x")
      assert(tostring(stmt.ifs[1].body[2])):eq("y")
      assert(tostring(stmt.ifs[2].conds[1])):eq("num")
      assert(stmt.ifs[2].body):len(2)
      assert(tostring(stmt.ifs[2].body[1])):eq("a")
      assert(tostring(stmt.ifs[2].body[2])):eq("b")
      assert(stmt.els):isNotNil()
      assert(stmt.els):len(2)
      assert(tostring(stmt.els[1])):eq("1")
      assert(tostring(stmt.els[2])):eq("2")
      assert(stmt.catch):isTable()
      assert(stmt.catch.var):eq("e")
      assert(stmt.catch.body):len(1)
      assert(stmt.catch.body[1]:__tostring()):eq("(call print y)")
      assert(stmt.finally):isTable()
      assert(stmt.finally.body):len(1)
      assert(stmt.finally.body[1]:__tostring()):eq("(call print z)")
      assert(parser:next()):isTable():has({
        line = 12,
        col = 1,
        type = StmtType.EXP
      })
    end)
  end):tags("with")
end):tags("stmt")
